import matplotlib.pyplot as plt
import numpy as np

def visualize(mat, true=None, size=0.75):
    size = np.sqrt(mat.shape[-1]) * size
    plt.rcParams['figure.figsize'] = [size if true is None else 2 * size, size]
    if true is None:
        fig, axs = plt.subplots(1, 1)
        mats, labels = [mat],  ["predicted"]
    else:
        fig, axs = plt.subplots(1, 2)
        mats, labels = [mat, true],  ["predicted", "true"]

    for ax, label, mat in zip(axs, labels, mats):
        ax.matshow(mat, vmin=0, vmax=1)
        plt.setp(ax.get_xticklabels(), visible=False)
        plt.setp(ax.get_yticklabels(), visible=False)
        ax.tick_params(axis='both', which='both', length=0)
        ax.set_title(label, pad=10)

    plt.tight_layout()
    plt.show()
    return




import math
import igraph as ig
import numpy as onp
import random as pyrandom
import pandas as pd

from avici.definitions import ROOT_DIR, GRAPH_YEAST, GRAPH_ECOLI
from avici.utils.graph import mat_to_graph, graph_to_mat
from avici.synthetic import GraphModel


def _module_extraction(rng, source, d, topk=0.2, at_least_n_regulators=1, make_acyclic=True, drop_self_edges=True,
                       return_node_indices=False):
    """
    Submodule extraction method by Marbach et al. (2009)
    https://www.liebertpub.com/doi/pdfplus/10.1089/cmb.2008.09TT

    Chooses a random node from the source network as seed and then adds
    a neighboring node (and all edges connecting the nodes in the source network) to the subnetwork randomly among the
    top k-percent "modular" neighbors.

    Important points:
      - The returned network contains all edges from the directed source network covered by the selected nodes
      - We consider the *undirected* skeleton of the source network for the modularity and neighborhood computation
        and ultimately selecting our subset of nodes
        (There is no definition of neighborhood/degree in the paper by Marbach et al (2009), which would require
        further explanation for a directed graph, and the original source code in GNW appears to only
        considers an undirected graph)

    Reference implementation:
    https://github.com/tschaffter/genenetweaver/blob/c5310349f5d5723306585c2bb62aedbdeb70db46/src/ch/epfl/lis/gnw/SubnetExtractor.java#L261
    https://github.com/tschaffter/jmod/blob/e2b230781ea3c3ad00833c58bd31b3b7d820317f/src/main/java/ch/epfl/lis/jmod/modularity/ModularityDetector.java#L204
    https://github.com/tschaffter/jmod/blob/e2b230781ea3c3ad00833c58bd31b3b7d820317f/src/main/java/ch/epfl/lis/jmod/JmodNetwork.java#L118
    (jmod is not the exact library used in genenetweaver, but is by the same author and satisfies almost the same api;
    I am using this as a reference since the actual imod library code used by genenetweaver is not public)

    Args:
        rng
        source: source network of shape (n_vars, n_vars)
        d: size of network we want to extract; d < n_vars
        topk: in [0.0, 1.0] percentage of neighbor nodes sorted by "modularity" from which we randomly choose the next node
        at_least_n_regulators: number of regulators/transcription factors to include with prob 1.
            A master regulator is a node with at least one outgoing edge. For the yeast graph used here, 
            157 of the 4284 nodes are master regulators/transcription factors
        make_acyclic
        drop_self_edges
        return_node_indices
    """

    dmax = source.shape[0]
    assert d < dmax
    assert d > at_least_n_regulators
    assert 0.0 <= topk <= 1.0
    assert source.shape[0] == source.shape[1]
    assert onp.all(onp.isclose(source, 0) | onp.isclose(source, 1))

    if drop_self_edges:
        source[onp.diag_indices(dmax)] = 0

    regulators = set(onp.where(source.sum(1) > 0)[0])

    # intialize modularity matrix B; do this based the undirected version of the source network (see docstring)
    a = source | source.T
    upper_tria_a = onp.triu(a, k=1)
    k = upper_tria_a.sum(0) + upper_tria_a.sum(1)
    m = upper_tria_a.sum()
    b = a - onp.outer(k, k) / (2 * m)
    assert m == 0.5 * k.sum()

    # randomly select seed
    s = -1 * onp.ones(dmax)
    if at_least_n_regulators:
        # if at_least_n_regulators > 0, then use a regulator here
        seed = rng.choice(list(regulators))
        assert source[seed, :].sum() > 0
    else:
        seed = rng.choice(dmax)

    s[seed] = 1
    subnet = {seed}

    # add d - 1 more nodes
    for _ in range(d - 1):

        # find neighbors of s in undirected source network
        neighbors = set()
        for i in subnet:
            for j in onp.where(a[i] == 1)[0]:
                if j not in subnet:
                    # if not enough regulators added yet, only add regulators to neighborhood
                    if at_least_n_regulators > 0 and len(regulators.intersection(subnet)) < at_least_n_regulators:
                        if j in regulators:
                            neighbors.add(j)
                    else:
                        neighbors.add(j)

        # if no neighbors (subnet is an island or we require a non-neighboring regulator), set all nodes as neighbors
        # in case we still require a regulator, this ensures we will add one this or next iteration
        if len(neighbors) == 0:
            neighbors = set(range(dmax)) - subnet

        assert len(neighbors.intersection(subnet)) == 0
        assert len(neighbors) > 0

        # compute modularity resulting from adding a neighbor
        qs, qs_nodes = [], []
        bterm_left = s @ b
        for neighbor in neighbors:
            assert s[neighbor] == -1
            assert neighbor not in subnet
            bterm_left_adj = bterm_left + 2 * b[neighbor, :]
            q_prop = (bterm_left_adj @ s) + 2 * bterm_left_adj[neighbor]
            qs.append(q_prop / (4 * m))
            qs_nodes.append(neighbor)

        # add random node among topk% modular graph options
        topk_q_indices = onp.argsort(qs)[::-1][:math.floor(topk * len(neighbors)) + 1]
        topk_q_nodes = onp.array(qs_nodes)[topk_q_indices]

        u = rng.choice(topk_q_nodes)
        assert s[u] == -1
        assert u not in subnet
        s[u] = 1
        subnet.add(u)

        assert set(onp.where(s == 1)[0]) == subnet

    # sanity checks once network extracted
    assert len(subnet) == d
    assert at_least_n_regulators == 0 or len(regulators.intersection(subnet)) >= at_least_n_regulators

    # extract directed subnetwork from source network based on selected node set
    subgraph = _extract_subnetwork(source, subnet)

    if make_acyclic:
        subgraph = _break_cycles_randomly(rng, subgraph)

    if return_node_indices:
        return subgraph, subnet
    else:
        return subgraph


def _extract_subnetwork(source, nodes):
    """
    Extracts subnetwork defindes by nodes from the source network

    Args:
        source: [n, n] adjacency matrix
        nodes: set of node indices in `source` defining the subnetwork
    """
    d = len(nodes)
    subgraph = onp.zeros((d, d), dtype=source.dtype)
    idx = {k: idx for idx, k in enumerate(nodes)}
    for node_i in nodes:
        for node_j in nodes:
            subgraph[idx[node_i], idx[node_j]] = source[node_i, node_j]

    return subgraph


def _break_cycles_randomly(rng, mat):
    """
    DFS that breaks cycles at random position through a random starting point
    """
    color = [0] * mat.shape[0]

    def dfs(u):
        color[u] = 1
        for v in onp.where(mat[u, :] == 1)[0]:
            if color[v] == 1:
                # back edge, which implies a cycle; remove edge that closes the cycle
                mat[u, v] = 0
            elif color[v] == 0:
                dfs(v)
        color[u] = 2

    for s in rng.permutation(mat.shape[0]):
        if color[s] == 0:
            dfs(s)

    assert mat_to_graph(mat).is_dag()
    return mat


def orient_pdag_randomly(rng, mat):
    """
    Orient PDAG randomly as a DAG by consistently orienting undirected edges with the partial ordering
    Done by viewing undirected edges as 2-cycles that are broken randomly
    """
    orig_mat = mat.copy()
    dag = _break_cycles_randomly(rng, mat)
    assert onp.all(~(((orig_mat == 1) & (orig_mat.T == 1)) & ((dag == 0) & (dag.T == 0)))), \
        "Some undirected edges were deleted completely"
    return dag


class ErdosRenyi(GraphModel):
    """
    Erdos-Renyi random graph
    
    Args:
        edges_per_var (float): expected number of edges per node, scaled to account for number of nodes
    """
    def __init__(self, edges_per_var):
        self.edges_per_var = edges_per_var

    def __call__(self, rng, n_vars):
        # select p s.t. we get requested edges_per_var in expectation
        n_edges = self.edges_per_var * n_vars
        p = min(n_edges / ((n_vars * (n_vars - 1)) / 2), 0.99)

        # sample
        mat = rng.binomial(n=1, p=p, size=(n_vars, n_vars)).astype(int) # bernoulli

        # make DAG by zeroing above diagonal; k=-1 indicates that diagonal is zero too
        dag = onp.tril(mat, k=-1)

        # randomly permute
        p = rng.permutation(onp.eye(n_vars).astype(int))
        dag = p.T @ dag @ p
        return dag


class ScaleFree(GraphModel):
    """
    Barabasi-Albert (scale-free)
    Power-law in-degree
    
    Args:
        edges_per_var (int): number of edges per node
        power (float): power in preferential attachment process. 
            Higher values make few nodes have high in-degree.

    """
    def __init__(self, edges_per_var, power=1.0):
        self.edges_per_var = edges_per_var
        self.power = power

    def __call__(self, rng, n_vars):
        pyrandom.seed(rng.bit_generator.state["state"]["state"]) # seed pyrandom based on state of numpy rng
        _ = rng.normal() # advance rng state by 1
        perm = rng.permutation(n_vars).tolist()
        g = ig.Graph.Barabasi(n=n_vars, m=self.edges_per_var, directed=True, power=self.power).permute_vertices(perm)
        mat = graph_to_mat(g)
        return mat


class ScaleFreeTranspose(ScaleFree):
    """
    Barabasi-Albert (scale-free)
    Power-law out-degree
    
    Args:
        edges_per_var (int): number of edges per node
        power (float): power in preferential attachment process. 
            Higher values make few nodes have high out-degree.

    """
    def __init__(self, edges_per_var, power=1):
        super().__init__(edges_per_var=edges_per_var, power=power)

    def __call__(self, rng, n_vars):
        mat = super().__call__(rng, n_vars)
        return mat.T


class WattsStrogatz(GraphModel):
    """
    Watts-Strogatz (small-world)

    Args:
        dim (int): the dimension of the lattice
        nei (int): value giving the distance (number of steps) within which two vertices will be connected.
        p (float): rewiring probability
    """
    def __init__(self, dim=2, nei=2, p=0.1):
        self.dim = dim
        self.nei = nei
        self.p = p

    def __call__(self, rng, n_vars):
        # choose size s.t. we get at smallest possible n_vars greater than requested n_vars given the dimension of lattice
        dim_size = math.ceil(n_vars ** (1.0 / self.dim))

        pyrandom.seed(rng.bit_generator.state["state"]["state"]) # seed pyrandom based on state of numpy rng
        _ = rng.normal() # advance rng state by 1
        g = ig.Graph.Watts_Strogatz(dim=self.dim, size=dim_size, nei=self.nei, p=self.p, multiple=False, loops=False)

        # drop excessive vertices s.t. we get exactly n_vars
        n_excessive = len(g.vs) - n_vars
        assert n_excessive >= 0
        if n_excessive:
            g.delete_vertices(rng.choice(g.vs, size=n_excessive, replace=False))
        assert len(g.vs) == n_vars, f"Didn't get requested graph; g.vs: {len(g.vs)}, n_vars {n_vars}"

        # make directed
        mat = onp.triu(graph_to_mat(g), k=1)

        # randomly permute
        p = rng.permutation(onp.eye(n_vars).astype(int))
        mat = p.T @ mat @ p

        assert ig.Graph.Weighted_Adjacency(mat.tolist()).is_dag()
        return mat


class SBM(GraphModel):
    """
    Stochastic Block Model

    Args:
        edges_per_var (int): expected number of edges per node
        n_blocks (int): number of blocks in model
        damp (float): if p is probability of intra block edges, damp * p is probability of inter block edges
            p is determined based on `edges_per_var`. For damp = 1.0, this is equivalent to erdos renyi
    """
    def __init__(self, edges_per_var, n_blocks, damp=0.2):
        self.edges_per_var = edges_per_var
        self.n_blocks = n_blocks
        self.damp = damp

    def __call__(self, rng, n_vars):
        # sample blocks
        splits = onp.sort(rng.choice(n_vars, size=self.n_blocks - 1, replace=False))
        blocks = onp.split(rng.permutation(n_vars), splits)
        block_sizes = onp.array([b.shape[0] for b in blocks])

        # select p s.t. we get requested edges_per_var in expectation
        block_edges_sampled = (onp.outer(block_sizes, block_sizes) - onp.diag(block_sizes)) / 2
        relative_block_probs = onp.eye(self.n_blocks) + self.damp * (1 - onp.eye(self.n_blocks))
        n_edges = self.edges_per_var * n_vars
        p = min(0.99, n_edges / onp.sum(block_edges_sampled * relative_block_probs))

        # sample graph
        mat_intra = rng.binomial(n=1, p=p, size=(n_vars, n_vars)).astype(int) # bernoulli
        mat_inter = rng.binomial(n=1, p=self.damp * p, size=(n_vars, n_vars)).astype(int) # bernoulli

        mat = onp.zeros((n_vars, n_vars))
        for i, bi in enumerate(blocks):
            for j, bj in enumerate(blocks):
                mat[onp.ix_(bi, bj)] = (mat_intra if i == j else mat_inter)[onp.ix_(bi, bj)]

        # make directed
        mat = onp.triu(mat, k=1)

        # randomly permute
        p = rng.permutation(onp.eye(n_vars).astype(int))
        mat = p.T @ mat @ p

        assert ig.Graph.Weighted_Adjacency(mat.tolist()).is_dag()
        return mat


class GRG(GraphModel):
    """
    Geometric random graph
    
    Args:
        radius (float): radius inside which random dots in a unit square are connected
    """
    def __init__(self, radius=0.1):
        self.radius = radius

    def __call__(self, rng, n_vars):
        pyrandom.seed(rng.bit_generator.state["state"]["state"])  # seed pyrandom based on state of numpy rng
        _ = rng.normal()  # advance rng state by 1
        g = ig.Graph._GRG(n=n_vars, radius=self.radius)[0]

        # make directed
        mat = onp.triu(graph_to_mat(g), k=1)

        # randomly permute
        p = rng.permutation(onp.eye(n_vars).astype(int))
        mat = p.T @ mat @ p

        assert ig.Graph.Weighted_Adjacency(mat.tolist()).is_dag()
        return mat


class Yeast(GraphModel):
    """
    Random subgraph extracted from real yeast graph
    using the extraction method https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.459.6733&rep=rep1&type=pdf
    
    Args:
        topk (float): heuristic in modularity extraction method introducing additional randomness.
            Percentage in [0.0, 1.0] of neighbor nodes sorted by "modularity" from which we randomly choose next node.
        at_least_n_regulators (int): number of regulators/transcription factors to include with prob 1.
             A master regulator is a node with at least one outgoing edge.
        make_acyclic (bool, optional): whether to make acyclic
    """
    def __init__(self, topk=0.2, at_least_n_regulators=1, make_acyclic=True):
        self.topk = topk
        self.at_least_n_regulators = at_least_n_regulators
        self.make_acyclic = make_acyclic

    def __call__(self, rng, n_vars):
        # load edges
        # (12873, 2)
        raw = pd.read_csv(ROOT_DIR / GRAPH_YEAST, sep="\t", index_col=False, header=None)

        # convert to numerical
        unique_genes = sorted(list(set(raw[0].unique().tolist() + raw[1].unique().tolist())))
        mapping = {k: v for v, k in enumerate(unique_genes)}
        edges = raw.applymap(lambda x: mapping[x]).to_numpy()

        g_raw = onp.zeros((len(unique_genes), len(unique_genes)), dtype=onp.int32) # shape (4441, 4441)
        g_raw[edges[:, 0], edges[:, 1]] = 1

        # extract random subnetwork
        g = _module_extraction(rng, g_raw, n_vars, topk=self.topk, at_least_n_regulators=self.at_least_n_regulators,
                               make_acyclic=self.make_acyclic, return_node_indices=False)

        # randomly permute
        p = rng.permutation(onp.eye(n_vars).astype(int))
        g_permuted = p.T @ g @ p
        return g_permuted


class Ecoli(GraphModel):
    """
    Random subgraph extracted from real e.coli graph
    using the extraction method https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.459.6733&rep=rep1&type=pdf
    
    Args:
        topk (float): heuristic in modularity extraction method introducing additional randomness.
            Percentage in [0.0, 1.0] of neighbor nodes sorted by "modularity" from which we randomly choose next node.
        at_least_n_regulators (int): number of regulators/transcription factors to include with prob 1.
             A master regulator is a node with at least one outgoing edge.
        make_acyclic (bool, optional): whether to make acyclic
    """
    def __init__(self, topk=0.2, at_least_n_regulators=1, make_acyclic=True):
        self.topk = topk
        self.at_least_n_regulators = at_least_n_regulators
        self.make_acyclic = make_acyclic

    def __call__(self, rng, n_vars):
        # load edges
        # (3758, 2)
        raw = pd.read_csv(ROOT_DIR / GRAPH_ECOLI, sep="\t", index_col=False, header=None)
        raw.loc[raw[2] == "+-", 2] = onp.nan
        raw.loc[raw[2] == "?", 2] = onp.nan

        # impute + or - randomly for +- and ?
        prob_pos = (raw[2] == "+").sum() / ((raw[2] == "+").sum() + (raw[2] == "-").sum())
        raw.loc[raw[2].isnull(), 2] = rng.choice(["+", "-"], p=[prob_pos, 1-prob_pos], size=raw[2].isnull().sum())
        assert set(raw[2].unique()) == {'+', '-'}

        # convert to numerical
        unique_genes = sorted(list(set(raw[0].unique().tolist() + raw[1].unique().tolist())))
        mapping = {k: v for v, k in enumerate(unique_genes)}
        edges = raw.loc[:, [0, 1]].applymap(lambda x: mapping[x]).to_numpy()
        effect_signs = raw.loc[:, [2]].applymap(lambda x: {"+": 1.0, "-": -1.0}[x]).to_numpy().squeeze(-1)
        assert onp.all(onp.isclose(effect_signs, -1.0) | onp.isclose(effect_signs, 1.0))

        g_raw = onp.zeros((len(unique_genes), len(unique_genes)), dtype=onp.int32) # shape (1565, 1565)
        effect_raw_sgn = onp.zeros((len(unique_genes), len(unique_genes)), dtype=onp.float32)
        g_raw[edges[:, 0], edges[:, 1]] = 1
        effect_raw_sgn[edges[:, 0], edges[:, 1]] = effect_signs

        # extract random subnetwork
        g, subnet = _module_extraction(rng, g_raw, n_vars, topk=self.topk, at_least_n_regulators=self.at_least_n_regulators,
                                       make_acyclic=self.make_acyclic, return_node_indices=True)

        effect_sgn = _extract_subnetwork(effect_raw_sgn, subnet)
        effect_sgn *= g.astype(onp.float32) # mask in case a cycle was broken

        # randomly permute
        p = rng.permutation(onp.eye(n_vars).astype(onp.int32))
        g_permuted = p.T @ g @ p
        # effect_sgn_permuted = p.T @ effect_sgn @ p
        return g_permuted



from avici.synthetic import Distribution


class Gaussian(Distribution):
    def __init__(self, scale=1.0):
        self.scale = scale

    def __call__(self, rng, shape=None):
        return self.scale * rng.normal(size=shape)


class Laplace(Distribution):
    def __init__(self, scale=1):
        self.scale = scale

    def __call__(self, rng, shape=None):
        return self.scale * rng.laplace(size=shape)


class Cauchy(Distribution):
    def __init__(self, scale=1.0):
        self.scale = scale

    def __call__(self, rng, shape=None):
        return self.scale * rng.standard_cauchy(size=shape)


class Uniform(Distribution):
    def __init__(self, low, high):
        self.low = low
        self.high = high

    def __call__(self, rng, shape=None):
        return rng.uniform(size=shape, low=self.low, high=self.high)


class SignedUniform(Distribution):
    def __init__(self, low, high):
        self.low = low
        self.high = high

    def __call__(self, rng, shape=None):
        sgn = rng.choice([-1, 1], size=shape)
        return sgn * rng.uniform(size=shape, low=self.low, high=self.high)


class RandInt(Distribution):
    def __init__(self, low, high, endpoint=True):
        self.low = low
        self.high = high
        self.endpoint = endpoint

    def __call__(self, rng, shape=None):
        return rng.integers(size=shape, low=self.low, high=self.high, endpoint=self.endpoint)


class Beta(Distribution):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __call__(self, rng, shape=None):
        return rng.beta(self.a, self.b, size=shape)



import igraph as ig
import numpy as onp

def graph_to_mat(g):
    """Returns adjacency matrix of ig.Graph object """
    return onp.array(g.get_adjacency().data).astype(int)

def mat_to_graph(mat):
    """Returns ig.Graph object for adjacency matrix """
    return ig.Graph.Weighted_Adjacency(mat.tolist())

def graph_to_toporder(g):
    """Returns adjacency matrix of ig.Graph object """
    return onp.array(g.topological_sorting()).astype(int)

def mat_to_toporder(mat):
    """Returns adjacency matrix of ig.Graph object """
    return onp.array(mat_to_graph(mat).topological_sorting()).astype(int)


import multiprocessing
import time
import psutil
from collections import defaultdict
import functools

import tensorflow as tf
import tensorflow_datasets as tfds
import jax
import numpy as onp

from avici.definitions import RNG_ENTROPY_TRAIN
from avici.buffer import SharedState, get_soft_sample_insert_ratio, is_buffer_filled
from avici.utils.tf import structured_py_function
import pyarrow.plasma as plasma

def _device_put_sharded(sharded_tree, devices):
    """Taken from  https://github.com/deepmind/dm-haiku/blob/main/examples/imagenet/dataset.py """
    leaves, treedef = jax.tree_util.tree_flatten(sharded_tree)
    n = leaves[0].shape[0]
    return jax.device_put_sharded([jax.tree_util.tree_unflatten(treedef, [l[i] for l in leaves]) for i in range(n)],
                                  devices)

def _double_cache(ds):
    """Taken from https://github.com/deepmind/dm-haiku/blob/main/examples/imagenet/dataset.py

    Keeps at least two batches on the accelerator.
    The current GPU allocator design reuses previous allocations. For a training
    loop this means batches will (typically) occupy the same region of memory as
    the previous batch. An issue with this is that it means we cannot overlap a
    host->device copy for the next batch until the previous step has finished and
    the previous batch has been freed.
    By double buffering we ensure that there are always two batches on the device.
    This means that a given batch waits on the N-2'th step to finish and free,
    meaning that it can allocate and copy the next batch to the accelerator in
    parallel with the N-1'th step being executed.
    Args:
    ds: Iterable of batches of numpy arrays.
    Yields:
    Batches of sharded device arrays.
    """
    batch = None
    devices = jax.local_devices()
    for next_batch in ds:
        assert next_batch is not None
        next_batch = _device_put_sharded(next_batch, devices)
        if batch is not None:
            yield batch
        batch = next_batch
    if batch is not None:
        yield batch


class AsyncBufferDataset:
    """
    Class for maintaining and sampling data batches from several buffers that are continually asynchronously updated

    If there are exceptions and errors after the main program exit()'s
    it is often because multiprocessing elements do not get cleaned up properly.
    Make sure pools, queues, processes are finished and emptied properly
    at the end of the main program.
    """

    def __init__(self, *,
                 seed,
                 config,
                 buffer_class,
                 buffer_size,
                 train_n_observations_obs,
                 train_n_observations_int,
                 batch_dims_train=None,
                 batch_dims_test=None,
                 double_cache_train=True,
                 prefetch=tf.data.experimental.AUTOTUNE,
                 n_workers=None,
                 n_listeners=1,
                 verbose=True,
                 queue_max_size=100,
                 object_store_gb=10.0,
                 ):

        self.config = config
        self.buffer_class = buffer_class
        self.train_n_observations_obs = train_n_observations_obs
        self.train_n_observations_int = train_n_observations_int
        self.batch_dims_train = batch_dims_train
        self.batch_dims_test = batch_dims_test
        self.double_cache_train = double_cache_train
        self.prefetch = prefetch
        self.verbose = verbose

        if self.verbose:
            print(f"AsyncBufferDataset", flush=True)

        # create plasma object store server
        # this emulates `with plasma.start_plasma_store as self._plasma_store: ...` and is closed in self.finish())
        self._plasma_store = plasma.start_plasma_store(int(object_store_gb * 1024 * 1024 * 1024))
        self._plasma_store_name, self._proc = type(self._plasma_store).__enter__(self._plasma_store)

        self._main_plasma_client = plasma.connect(self._plasma_store_name)

        # setup workers via async multiprocessing
        try:
            # only on linux but only correct call on cluster (psutil.cpu_count is false)
            cpu_count = len(psutil.Process().cpu_affinity())
        except AttributeError:
            cpu_count = psutil.cpu_count(logical=True)

        self.n_workers = n_workers or cpu_count
        self.n_listeners = n_listeners if self.n_workers >= 16 else 1 # with few CPUs don't need more than one listener

        n_processes = self.n_workers + self.n_listeners + 1
        self._pool = multiprocessing.Pool(processes=n_processes)
        self._endless_jobs = []
        self._listener_jobs = []
        if self.verbose:
            print(f"cpu count: {cpu_count}", flush=True)
            print(f"processes: {n_processes} "
                  f"({self.n_workers} workers,"
                  f" {self.n_listeners} listeners,"
                  f" {1} load balancer)", flush=True)

        # init shared state
        manager = multiprocessing.Manager()

        self._state = SharedState(
            queues=[manager.Queue(maxsize=queue_max_size) for _ in range(self.n_listeners)],
            buffers=defaultdict(dict),
            is_alive=manager.Event(),
            next_job=manager.Value(int, self.config["train_n_vars"][0]),
            plasma_store_name=self._plasma_store_name,
        )
        self._state.is_alive.set()

        # buffers
        for descr, spec in self.config["data"].items():
            is_train = descr == "train"
            for n_vars in self.config["train_n_vars"] if is_train else self.config["test_n_vars"]:
                size = buffer_size if is_train else self.config["test_n_datasets"]
                self._state.buffers[descr][n_vars] = self.buffer_class.init(manager, size)

        # launch workers that fill the buffers
        self._launch(seed)


    """Buffer"""

    def get_qsize(self):
        return [q.qsize() for q in self._state.queues]

    def get_max_size(self):
        return self._tree_map(lambda b: b.max_size, self._state.buffers)

    def get_size(self):
        return self._tree_map(lambda b: b.state.size, self._state.buffers)

    def get_items_added(self):
        return self._tree_map(lambda b: b.metrics.items_added, self._state.buffers)

    def get_items_sampled(self):
        return self._tree_map(lambda b: b.metrics.items_sampled, self._state.buffers)

    def get_soft_sample_insert_ratio(self):
        return self._tree_map(get_soft_sample_insert_ratio, self._state.buffers["train"])

    def get_is_filled(self):
        return self._tree_map(is_buffer_filled, self._state.buffers)

    def log_scalars_dict(self, aux=None, steps=1):
        assert steps > 0
        items_added_ = self.get_items_added()["train"]
        items_sampled_ = self.get_items_sampled()["train"]
        sample_insert_ratio_ = self.get_soft_sample_insert_ratio()

        # init running counts of items added and sampled
        if aux is None:
            aux = {}
            aux.update({f"added_d={n_vars}": 0 for n_vars in items_added_.keys()})
            aux.update({f"sampled_d={n_vars}": 0 for n_vars in items_sampled_.keys()})

        # make scalars
        d = {
            f"buffer/qsize": sum(self.get_qsize()) / len(self._state.queues),
            f"buffer/plasma_store": len(self._main_plasma_client.list()),
        }
        for n_vars in items_added_.keys():
            d.update({f"buffer/added_per_step-d={n_vars}": (items_added_[n_vars] - aux[f"added_d={n_vars}"]) / steps})
            d.update({f"buffer/sampled_per_step-d={n_vars}": (items_sampled_[n_vars] - aux[f"sampled_d={n_vars}"]) / steps})
            d.update({f"buffer/sample_insert_ratio-d={n_vars}": sample_insert_ratio_[n_vars]})

            # update aux counts
            aux[f"added_d={n_vars}"] = items_added_[n_vars]
            aux[f"sampled_d={n_vars}"] = items_sampled_[n_vars]

        return d, aux

    def _tree_map(self, f, tree, *args, **kwargs):
        """Apply f to every leaf and return result"""
        if type(tree) in [dict, defaultdict, multiprocessing.managers.DictProxy]:
            return {k: self._tree_map(f, subtree, *args, **kwargs) for k, subtree in tree.items()}
        else:
            return f(tree, *args, **kwargs)

    def _flatten(self, tree):
        """Flatten nested dict into list"""
        if type(tree) in [dict, defaultdict, multiprocessing.managers.DictProxy]:
            for _, subtree in tree.items():
                yield from self._flatten(subtree)
        else:
            yield tree


    """Async multiprocessing"""

    def _launch_async_workers(self, descr, n_vars, seed, size=None, endless=False):
        """launches async workers that fill buffer `buffer_id`"""
        assert endless or (size is not None)
        n_workers_needed = self.n_workers if size is None else min(self.n_workers, size)
        worker_seeds = seed.spawn(n_workers_needed)
        jobs = []
        for j, worker_seed in enumerate(worker_seeds):
            it = None if endless else range(j, size, n_workers_needed)
            p = self._pool.apply_async(
                self.buffer_class.async_worker,
                (j, self._state, worker_seed, it, descr, n_vars, self.config))

            p.daemon = True
            jobs.append(p)

        return jobs


    def _launch(self, seed, t_print=20.0):
        """
        Launch producer pipeline. Fill all buffers once and wait until done. Then launch
        endless workers for train buffers
        """
        if self.verbose:
            print(f"Filling all buffers...", flush=True)

        self.ds_seed, loop_seed, *init_seeds = \
            onp.random.SeedSequence(entropy=(RNG_ENTROPY_TRAIN, seed)).spawn(len(self._state.buffers) + 2)

        # put load balancer to work
        plb = self._pool.apply_async(self.buffer_class.load_balanc_async, (self._state,))
        plb.daemon = True
        self._endless_jobs.append(plb)

        # put listeners to work (filling the buffer with produced data on queues)
        for j in range(self.n_listeners):
            pli = self._pool.apply_async(self.buffer_class.insert_async, (j, self._state))
            pli.daemon = True
            self._listener_jobs.append(pli)

        # sequentially fill all buffers using `n_workers` workers that produce for the queues
        init_jobs = []
        for (descr, buffer_dict), seed in zip(self._state.buffers.items(), init_seeds):
            for n_vars, buffer in buffer_dict.items():
                init_jobs += self._launch_async_workers(descr, n_vars, seed, size=buffer.max_size, endless=False)

        # wait for buffers to be full
        t_ref = time.time()
        while True and self.verbose:
            if all(list(self._flatten(self.get_is_filled()))):
                break
            if time.time() > t_ref + t_print:
                sizes, max_sizes = self.get_size(), self.get_max_size()
                strg = f"buffers: "
                for descr, buffer_dict in self._state.buffers.items():
                    strg += f"{descr}: "
                    for n_vars, _ in buffer_dict.items():
                        strg += f"{sizes[descr][n_vars]:>3d}/{max_sizes[descr][n_vars]}  "
                print(strg, flush=True)
                t_ref = time.time()

                # check for raised exceptions in workers
                try:
                    if any([not job.successful() for job in init_jobs]):
                        break
                except ValueError as e:
                    if "multiprocessing.pool.ApplyResult" in str(e) and "not ready" in str(e):
                        pass  # job is not done yet
                    else:
                        raise e

        for job in init_jobs:
            _ = job.get()  # if verbose == True, returns immediately
        if self.verbose:
            print(f"Buffers filled.", flush=True)
            
        # launch endless workers for train buffers
        self._endless_jobs += self._launch_async_workers("train", None, loop_seed, endless=True)

        if self.verbose:
            print(f"Launched endless background workers for train buffers.", flush=True)


    def finish(self):
        # do not attempt to change the ordering of these calls
        self._state.is_alive.clear()

        # wait to collect results from the train workers through the pool result queue
        for job in self._endless_jobs:
            _ = job.get()

        # when here, all async workers exited and we are done, so kill the listeners
        for j in range(self.n_listeners):
            self._state.queues[j].put('kill')
        for job in self._listener_jobs:
            _ = job.get()

        # clear the queue to avoid any post-exit errors when the script finishes
        for j in range(self.n_listeners):
            while not self._state.queues[j].empty():
                self._state.queues[j].get()
            self._state.queues[j].join()

        self._pool.close()
        self._pool.terminate()
        self._pool.join()

        # close plasma server
        # this emulates the end of indented `with` block (see when initialized)
        self._main_plasma_client.disconnect()
        type(self._plasma_store).__exit__(self._plasma_store, None, None, None)

        if self.verbose:
            print("Terminated AsyncBufferDataset", flush=True)

    """Dataset"""

    def _make_dataset(self, buffer, is_train: bool, batch_dims=None):
        """
        Launch consumer (buffer consumer) tf.data.Dataset iterator to support batching, prefetching, etc.
        for buffers `descr`
        """

        if is_train:
            # plasma client
            client = plasma.connect(self._plasma_store_name)

            # infinite dataset that samples continually from updating buffer
            ds = tf.data.Dataset.from_tensor_slices([0]).repeat(None)  # loop indefinitely
            rng = onp.random.default_rng(self.ds_seed)

            f = functools.partial(self.buffer_class.uniform_sample_from_buffer,
                                  rng=rng,
                                  client=client,
                                  buffer=buffer,
                                  n_obs=self.train_n_observations_obs + self.train_n_observations_int,
                                  n_int=self.train_n_observations_int)

        else:
            # dataset is simply the buffer
            ds = tf.data.Dataset.from_tensor_slices(list(range(buffer.max_size)))

            f = functools.partial(self.buffer_class.index_into_buffer, client=self._main_plasma_client, buffer=buffer)

        ds = ds.map(lambda *args: structured_py_function(func=f,
                                                         inp=[*args],
                                                         Tout=self.buffer_class.dtype),
                    # num_parallel_calls is super important once we have multiple devices
                    # on 4 gpus, empirically this gave 10x data loading performance
                    # (0.05 sec instead of 0.5sec for next(iter) in train loop)
                    num_parallel_calls=tf.data.experimental.AUTOTUNE)

        # create batches
        if batch_dims is not None:
            for i, batch_size in enumerate(reversed(batch_dims)):
                ds = ds.batch(batch_size)

        ds = ds.prefetch(self.prefetch)
        ds = tfds.as_numpy(ds)

        if self.double_cache_train and is_train:
            ds = _double_cache(ds)

        yield from ds


    def make_datasets(self, descr):
        """Generate dict of datasets {n_vars: tf.data.Dataset} for config `descr`"""
        ds = {}
        is_train = descr == "train"
        for n_vars, buffer in self._state.buffers[descr].items():
            batch_dims = self.batch_dims_train[n_vars]["device"] if is_train else self.batch_dims_test[n_vars]["device"]
            ds[n_vars] = self._make_dataset(buffer, is_train=is_train, batch_dims=batch_dims)

        return ds

    def make_test_datasets(self):
        test_set_ids = [k for k in self.config["data"].keys() if k != "train"]
        return {descr: self.make_datasets(descr) for descr in test_set_ids}



from pathlib import Path

# rng entropies; these integers must be different to guarantee different randomness in data during train, val, test
RNG_ENTROPY_TRAIN = 0
RNG_ENTROPY_TEST = 1
RNG_ENTROPY_HPARAMS = 2

# directories
ROOT_DIR = Path(__file__).parents[0]
PROJECT_DIR = Path(__file__).parents[1]

CHECKPOINT_SUBDIR = "checkpoints"
ASSETS_SUBDIR = "assets"
CACHE_SUBDIR = "cache"
SERGIO_NOISE_CONFIG = ROOT_DIR / "synthetic/sergio/noise_config.yaml"

# real data
REAL_DATASET_KEY = "real_dataset"
GRAPH_YEAST = "assets/yeast_transcriptional_network_Balaji2006.tsv"
GRAPH_ECOLI = "assets/ecoli_transcriptional_network_regulonDB_6_7.tsv"

# yaml
CHECKPOINT_KWARGS = "kwargs.json"
YAML_CLASS = "__class__"
YAML_MODULES = "additional_modules"

# huggingface
HUGGINGFACE_REPO = "larslorch/avici"
MODEL_NEURIPS_LINEAR = [
    "neurips-linear/kwargs.json",
    "neurips-linear/checkpoint_0300000.pkl",
]
MODEL_NEURIPS_RFF = [
    "neurips-rff/kwargs.json",
    "neurips-rff/checkpoint_0300000.pkl",
]
MODEL_NEURIPS_GRN = [
    "neurips-grn/kwargs.json",
    "neurips-grn/checkpoint_0300000.pkl",
]
MODEL_NEURIPS_SCM_V0 = [
    "scm-v0/kwargs.json",
    "scm-v0/checkpoint_0300000.pkl",
]



import warnings
warnings.formatwarning = lambda msg, category, path, lineno, file: f"{path}:{lineno}: {category.__name__}: {msg}\n"
warnings.filterwarnings("ignore", message="invalid value encountered in divide")
warnings.filterwarnings("ignore", message="divide by zero encountered in log2")

import numpy as onp

"""Standardization"""

def _onp_cpm_standardizer(x):
    # compute library sizes (sum of row)
    x_libsize = x.sum(-1, keepdims=True)

    # divide each cell by library size and multiply by 10^6
    # will yield nan for rows with zero expression and for zero expression entries
    log2cpm = onp.where(onp.isclose(x, 0.0), onp.nan, onp.log2(x / (x_libsize * 1e-6)))
    return log2cpm


def _onp_cpm_shift_scale(x, shift, scale):
    # shift and scale
    x = (x - onp.where(onp.isnan(shift), 0.0, shift)) / onp.where(onp.isnan(scale), 1.0, scale)

    # set nans (set for all-zero rows, i.e. zero libsize) to minimum (i.e. to zero)
    # catch empty arrays (when x_int is empty and has axis n=0)
    if not x.size == 0:
        x = onp.where(onp.isnan(x), 0.0, x)

    return x


def _onp_standardize_count(x_obs, x_int, x_obs_ho, x_int_ho):
    """
    log2 CPM normalization for gene expression count data
    https://bioconductor.org/packages/release/bioc/vignettes/edgeR/inst/doc/edgeRUsersGuide.pdf
    https://rdrr.io/bioc/edgeR/src/R/cpm.R
    http://luisvalesilva.com/datasimple/rna-seq_units.html

    log2 CPM(mat) = log2( 10^6 * mat / libsize ) where libsize = mat.sum(1) (i.e. sum over genes)

    Why log2(1+cpm) is a bad idea https://support.bioconductor.org/p/107719/
    Specific scaling https://support.bioconductor.org/p/59846/#59917

    """
    # cpm normalization
    x_obs[..., 0] = _onp_cpm_standardizer(x_obs[..., 0])
    x_int[..., 0] = _onp_cpm_standardizer(x_int[..., 0])

    # subtract min (~robust global median) and divide by global std dev
    global_ref_x = onp.concatenate([x_obs[..., 0], x_int[..., 0]], axis=-2)
    global_min = onp.nanmin(global_ref_x, axis=(-1, -2), keepdims=True)
    global_std = onp.nanstd(global_ref_x, axis=(-1, -2), keepdims=True)

    x_obs[..., 0] = _onp_cpm_shift_scale(x_obs[..., 0], global_min, global_std)
    x_int[..., 0] = _onp_cpm_shift_scale(x_int[..., 0], global_min, global_std)

    if x_obs_ho is not None and x_int_ho is not None:
        x_obs_ho[..., 0] = _onp_cpm_standardizer(x_obs_ho[..., 0])
        x_int_ho[..., 0] = _onp_cpm_standardizer(x_int_ho[..., 0])
        x_obs_ho[..., 0] = _onp_cpm_shift_scale(x_obs_ho[..., 0], global_min, global_std)
        x_int_ho[..., 0] = _onp_cpm_shift_scale(x_int_ho[..., 0], global_min, global_std)

    return x_obs, x_int, x_obs_ho, x_int_ho


def _onp_standardize_default(x_obs, x_int, x_obs_ho, x_int_ho):
    # default z-standardization
    ref_x = onp.concatenate([x_obs[..., 0], x_int[..., 0]], axis=-2)
    mean = ref_x.mean(-2, keepdims=True)
    std = ref_x.std(-2, keepdims=True)
    x_obs[..., 0] = (x_obs[..., 0] - mean) / onp.where(std == 0.0, 1.0, std)
    x_int[..., 0] = (x_int[..., 0] - mean) / onp.where(std == 0.0, 1.0, std)

    if x_obs_ho is not None and x_int_ho is not None:
        x_obs_ho[..., 0] = (x_obs_ho[..., 0] - mean) / onp.where(std == 0.0, 1.0, std)
        x_int_ho[..., 0] = (x_int_ho[..., 0] - mean) / onp.where(std == 0.0, 1.0, std)

    return x_obs, x_int, x_obs_ho, x_int_ho


def onp_standardize_data(data):
    """
    Standardize data pytree
    Heldout data is standardized using the inference data statistics
    """
    x_obs = data["x_obs"]
    x_int = data["x_int"]

    x_obs_ho = data.get("x_heldout_obs", None)
    x_int_ho = data.get("x_heldout_int", None)

    assert (x_obs.shape[-1] == 1 or x_obs.shape[-1] == 2) \
       and (x_int.shape[-1] == 1 or x_int.shape[-1] == 2), \
        f"Assume concat 3D shape but got: x_obs {x_obs.shape} and x_int {x_int.shape}"

    if data["is_count_data"]:
        x_obs, x_int, x_obs_ho, x_int_ho = _onp_standardize_count(x_obs, x_int, x_obs_ho, x_int_ho)
    else:
        x_obs, x_int, x_obs_ho, x_int_ho = _onp_standardize_default(x_obs, x_int, x_obs_ho, x_int_ho)

    if x_obs_ho is not None and x_int_ho is not None:
        return {**data, "x_obs": x_obs, "x_int": x_int, "x_heldout_obs": x_obs_ho, "x_heldout_int": x_int_ho}
    else:
        return {**data, "x_obs": x_obs, "x_int": x_int}


"""Data batching"""

def onp_subbatch(rng, data, n_obs, n_int):
    """Sample random subbatch from observations `x_obs` and `x_int` in data"""
    x_obs = data["x_obs"]
    x_int = data["x_int"]
    assert "x_heldout_obs" not in data and "x_heldout_int" not in data

    if n_obs > x_obs.shape[-3]:
        warnings.warn(f"Warning: Larger batch requested than available in subbatch. "
                      f"Got `n_obs`={n_obs} and `x_obs` shape {x_obs.shape}. Will just use `x_obs`")

    if n_int > x_int.shape[-3]:
        warnings.warn(f"Warning: Larger batch requested than available in subbatch. "
                      f"Got `n_int`={n_int} and `x_int` shape {x_int.shape}. Will just use `x_int`")

    if rng is None:
        x_obs = x_obs[..., :n_obs, :, :]
        x_int = x_int[..., :n_int, :, :]
    else:
        x_obs = x_obs[..., rng.permutation(x_obs.shape[-3])[:n_obs], :, :]
        x_int = x_int[..., rng.permutation(x_int.shape[-3])[:n_int], :, :]

    data = {**data, "x_obs": x_obs, "x_int": x_int}
    return data


def onp_bootstrap(rng, data):
    """Sample bootstrap data set from `data`"""
    x_obs = data["x_obs"]
    x_int = data["x_int"]
    n_total = x_obs.shape[0] + x_int.shape[0]
    assert "x_heldout_obs" not in data and "x_heldout_int" not in data

    bootstrap = rng.choice(n_total, size=n_total)
    bootstrap_obs = bootstrap[bootstrap < x_obs.shape[0]]
    bootstrap_int = bootstrap[bootstrap >= x_obs.shape[0]] - x_obs.shape[0]
    x_obs = x_obs[..., bootstrap_obs, :, :]
    x_int = x_int[..., bootstrap_int, :, :]
    assert x_obs.shape[0] + x_int.shape[0] == data["x_obs"].shape[0] + data["x_int"].shape[0], \
        f"The total number of observations in bootstrap data set changed. " \
        f"{x_obs.shape} {x_int.shape}  {data['x_obs'].shape} {data['x_int'].shape} "

    data = {**data, "x_obs": x_obs, "x_int": x_int}
    return data





from .abstract import Distribution, GraphModel, MechanismModel, NoiseModel, SyntheticSpec, CustomClassWrapper, Data
from .distributions import Gaussian, Laplace, Cauchy, Uniform, SignedUniform, RandInt, Beta
from .graph import ErdosRenyi, ScaleFree, ScaleFreeTranspose, WattsStrogatz, SBM, GRG, Yeast, Ecoli
from .noise_scale import SimpleNoise, HeteroscedasticRFFNoise
from .linear import LinearAdditive
from .rff import RFFAdditive
from .gene import GRNSergio



from avici.synthetic.noise_scale import init_noise_dist
from avici.synthetic.utils import sample_recursive_scm
from avici.synthetic import MechanismModel


class LinearAdditive(MechanismModel):
    """
    Linear mechanism with additive noise

    Args:
        param (Distribution): distribution for sampling linear function weights.
            Example: `avici.synthetic.SignedUniform` (for bounding away from zero)
        bias (Distribution): distribution for sampling linear function biases. Example: `avici.synthetic.Uniform`
        noise (Distribution): distribution for sampling additive noise. Example: `avici.synthetic.Gaussian`
        noise_scale (Distribution): if specified, distribution used to sample noise scale
            for `avici.synthetic.SimpleNoise`. Example: `avici.synthetic.Uniform`.
            Only one of `noise_scale` and `noise_scale_heteroscedastic` must be specified.
        noise_scale_heteroscedastic (Distribution): if specified, kwargs for
            initalizing `avici.synthetic.HeteroscedasticRFFNoise`.
            Only one of `noise_scale` and `noise_scale_heteroscedastic` must be specified.
        n_interv_vars (int): no. unique variables intervened upoin in all of data collected;
            -1 indicates all nodes
        interv_dist (Distribution): distribution for sampling intervention values.
            Example: `avici.synthetic.SignedUniform`
    """
    def __init__(self,
                 param,
                 bias,
                 noise,
                 noise_scale=None,
                 noise_scale_heteroscedastic=None,
                 n_interv_vars=0,
                 interv_dist=None):

        assert interv_dist is not None or n_interv_vars == 0

        self.param = param
        self.bias = bias
        self.noise = noise
        self.noise_scale = noise_scale
        self.noise_scale_heteroscedastic = noise_scale_heteroscedastic
        self.n_interv_vars = n_interv_vars
        self.interv_dist = interv_dist

    def __call__(self, rng, g, n_observations_obs, n_observations_int):

        assert self.interv_dist is not None or self.n_interv_vars == 0

        # construct mechanism for each node
        n_vars = g.shape[-1]
        f = []
        for j in range(n_vars):
            # sample parameters
            # each call may use random hyperparameters, so call once per node
            w = self.param(rng, shape=(n_vars,))
            b = self.bias(rng, shape=(1,))

            # bind parameters to mechanism function
            f.append(lambda x, is_parent, z, theta=w, bias=b: (x @ (theta * is_parent)) + bias + z)

        # construct noise distribution for each node
        nse = []
        for j in range(n_vars):
            # sample parameters and bind to sampling function
            # each call may sample a random noise scale, so call once per node
            nse.append(init_noise_dist(rng=rng,
                                       dim=int(g[:, j].sum().item()),
                                       dist=self.noise,
                                       noise_scale=self.noise_scale,
                                       noise_scale_heteroscedastic=self.noise_scale_heteroscedastic))

        # sample recursively over g given functionals and endogenous noise distribution
        data = sample_recursive_scm(
            rng=rng,
            n_observations_obs=n_observations_obs,
            n_observations_int=n_observations_int,
            g=g,
            f=f,
            nse=nse,
            interv_dist=self.interv_dist,
            n_interv_vars=self.n_interv_vars,
        )
        return data


import tensorflow as tf

"""
Workaround for allowing `tf.py_function` that returns structured elements like dict or NamedTuple
Taken from 
https://github.com/tensorflow/tensorflow/issues/27679#issuecomment-522578000
"""

def structured_py_function(func, inp, Tout, name=None):
    def wrapped_func(*flat_inp):
        reconstructed_inp = tf.nest.pack_sequence_as(inp, flat_inp,
                                                     expand_composites=True)
        result = func(*reconstructed_inp)
        return tf.nest.flatten(result, expand_composites=True)
    flat_Tout = tf.nest.flatten(Tout, expand_composites=True)
    flat_out = tf.py_function(
        func=wrapped_func,
        inp=tf.nest.flatten(inp, expand_composites=True),
        Tout=[_tensor_spec_to_dtype(v) for v in flat_Tout],
        name=name)
    spec_out = tf.nest.map_structure(_dtype_to_tensor_spec, Tout,
                                   expand_composites=True)
    out = tf.nest.pack_sequence_as(spec_out, flat_out, expand_composites=True)
    return out

def _dtype_to_tensor_spec(v):
    return tf.TensorSpec(None, v) if isinstance(v, tf.dtypes.DType) else v

def _tensor_spec_to_dtype(v):
    return v.dtype if isinstance(v, tf.TensorSpec) else v


import math
import functools
import time
from collections import defaultdict

import jax
import jax.random as random
import jax.numpy as jnp
import pandas as pd

from avici.utils.data import onp_subbatch
from avici.utils.data_jax import jax_get_x

from avici.metrics import shd, threshold_metrics, is_cyclic, classification_metrics, n_edges


def print_header(strg):
    print("\n" + "=" * 25 + f" {strg} " + "=" * 25, flush=True)


def update_ave(ave_d, d, n=1, is_mean=False):
    for k, v in d.items():
        ave_d[("__ctr__", k)] += n
        if is_mean:
            ave_d[k] += v * n
        else:
            ave_d[k] += v
    return ave_d


def retrieve_ave(ave_d):
    out = {}
    for k, v in ave_d.items():
        # check if `k` is a ctr element
        if isinstance(k, tuple) and k[0] == "__ctr__":
            continue
        # process value `v`
        try:
            v_val = v.item()
        # distributed training case with `pmap`: retrieve 1 device replicate
        except TypeError:
            v_val = v[0].item()
        # not an array
        except AttributeError:
            v_val = v
        assert ("__ctr__", k) in ave_d.keys()
        out[k] = v_val / ave_d[("__ctr__", k)]
    return out


@functools.partial(jax.jit, static_argnums=(0,))
def eval_batch_helper(model, params, dual, batch, key, step, decision_thresholds=(0.5,)):
    """jit-forward passes of net on a batch"""

    # loss
    key, subk = random.split(key)
    loss = model.loss(params, dual, subk, batch, step, False)[0]

    # graph preds
    x = jax_get_x(batch)
    g_edges_prob = model.infer_edge_probs(params, x)
    g_edges = {f"{thres}".replace(".", "_"): (g_edges_prob > thres).astype(jnp.int32) for thres in decision_thresholds}

    return {
        f"loss": loss,
        f"dual": dual,
        f"g_edges_prob": g_edges_prob,
        f"g_edges": g_edges,
    }


def evaluate(model, state, ds, step, detailed_log=False, log_metrics=("auprc", "auroc", "f1_0_5", "cyclic_0_5")):

    """Eval model on a full tree of data iterators at the given params/state"""

    results = {}
    df = pd.DataFrame()
    all_test_iters = ds.make_test_datasets()
    if len(all_test_iters) == 0:
        return {}

    # evaluate each test iterator
    for descr, test_iters in all_test_iters.items():
        for n_vars, it in test_iters.items():

            scalars_aux_stratified = defaultdict(lambda: defaultdict(float))
            t0 = time.time()
            key = state.rng

            # get batch of observations
            main_batch_size = None
            for batch in it:

                assert batch["x_obs"].ndim == 4 and batch["x_int"].ndim == 4 and batch["g"].ndim == 3

                # avoid recompilation on last remainder batch
                batch_size = batch["g"].shape[0] # size of batch
                if main_batch_size is not None and batch_size != main_batch_size:
                    break
                else:
                    main_batch_size = batch_size

                # evaluate batch for variable number of observations
                ratio_obs_int = batch["n_observations_obs"][0] / (batch["n_observations_obs"][0] + batch["n_observations_int"][0])
                n_obs_max = batch["n_observations_obs"][0] + batch["n_observations_int"][0]
                n_obs_l = [i for i in [50, 200, 1000] if i < n_obs_max] + [n_obs_max]

                # careful: every n_obs needs 1 compile
                for n in n_obs_l:

                    n_obs, n_int = math.floor(n * ratio_obs_int), math.ceil(n * (1 - ratio_obs_int))
                    subbatch = onp_subbatch(None, batch, n_obs, n_int)

                    scalars = defaultdict(float)

                    # predictions for normal checkpoint and for polyak checkpoint
                    for suffix, params in [("", state.params), ("_polyak", state.ave_params)]:

                        key, subk = random.split(key)
                        preds = eval_batch_helper(model, params, state.dual, subbatch, subk, step)

                        # loss
                        scalars.update({f"loss{suffix}": preds["loss"]})

                        for j in range(batch_size):

                            # shd
                            # at thresholds
                            for thres, g_edges in preds["g_edges"].items():
                                scalars[f"shd_{thres}{suffix}"] += \
                                    shd(subbatch["g"][j], g_edges[j]) / batch_size

                            # cyclicity
                            # at thresholds
                            for thres, g_edges in preds["g_edges"].items():
                                scalars[f"edges_{thres}{suffix}"] += \
                                    n_edges(g_edges[j]) / batch_size
                                scalars[f"cyclic_{thres}{suffix}"] += \
                                    float(is_cyclic(g_edges[j])) / batch_size

                            # precision metrics
                            # at thresholds
                            for thres, g_edges in preds["g_edges"].items():
                                metrics = classification_metrics(subbatch["g"][j], g_edges[j])
                                for k, v in metrics.items():
                                    scalars[f"{k}_{thres}{suffix}"] += v / batch_size

                            #probabilistic metrics
                            metrics = threshold_metrics(subbatch["g"][j], preds["g_edges_prob"][j])
                            for k, v in metrics.items():
                                scalars[f"{k}{suffix}"] += v / batch_size

                    ## save scalars
                    scalars_aux_stratified[n] = update_ave(scalars_aux_stratified[n], scalars, n=batch_size, is_mean=True)

            eval_time = time.time() - t0

            # compute averages and log scalars to df
            scalars_ave_stratified = {i: retrieve_ave(aux) for i, aux in scalars_aux_stratified.items()}
            for n, scalars_ave in scalars_ave_stratified.items():
                df = df.append({
                    **scalars_ave,
                    "descr": descr,
                    "d": n_vars,
                    "n": n,
                    "eval_time": eval_time,
                }, ignore_index=True)


    # collect results for max number of observations
    metrics_strs = df.columns.drop(["descr", "n", "d"], errors="ignore")
    if log_metrics is not None and not detailed_log:
        metrics_strs = metrics_strs\
            .drop([s for s in metrics_strs if not any([s in k for k in log_metrics])], errors="ignore") \
            .drop([s for s in metrics_strs if "polyak" in s], errors="ignore")

    for descr in df["descr"].unique():
        df_descr = df.loc[df["descr"] == descr]
        n_max = int(df_descr["n"].max())
        d_min = int(df_descr["d"].min())
        d_max = int(df_descr["d"].max())
        for d in df_descr["d"].unique():
            if not detailed_log and d not in [d_min, d_max]:
                continue
            results_train_settings = df_descr[(df_descr["n"] == n_max) & (df_descr["d"] == d)]
            assert not results_train_settings.empty, f"No results for {descr} d={d} n_max={n_max}"
            for metric_str in metrics_strs:
                results.update({f"{descr}/d={d}/{metric_str}": results_train_settings[metric_str].item()})

    return results



import math
import numpy as onp
import copy
from collections import defaultdict

from avici.synthetic.sergio.sim import Sergio as SergioSimulator
from avici.utils.load import load_yaml
from avici.definitions import SERGIO_NOISE_CONFIG
from avici.synthetic import MechanismModel, Data


class GRNSergio(MechanismModel):
    """
    SERGIO simulator for GRNs

    Args:
        b (Distribution): distribution for sampling basic reproduction rates. Example: `avici.synthetic.Uniform`
        k_param (Distribution): distribution for sampling (non-negative) interaction strenghts.
            Example: `avici.synthetic.Uniform`
        k_sign_p (Distribution): distribution of sampling probability for positive (vs. negative)
            interaction sign signs. Example: `avici.synthetic.Beta`
        hill (float): Hill function coefficient
        decays (float): decay rate
        noise_params (float): noise scale parameter
        cell_types (Distribution): distribution for sampling integer number of cell types.
            Example: `avici.synthetic.RandInt`
        noise_type (str): noise type in SERGIO simulator. Default: `dpd`
        sampling_state (int): configuration of SERGIO sampler. Default: 15
        dt (float): dt increment in stochastic process. Default: 0.01

        * Technical noise*

        tech_noise_config (str): specification of noise elvels.
            Select one of the keys in `avici/synthetic/sergio/noise_config.yaml`
        add_outlier_effect (bool): whether to simulate outlier effects based on `tech_noise_config`
        add_lib_size_effect (bool): whether to simulate library size effects based on `tech_noise_config`
        add_dropout_effect (bool): whether to simulate dropout effects based on `tech_noise_config`
        return_count_data (bool): whether to return Poisson count data of the float mean expression levels

        * Interventions *

        n_ko_genes (int): no. unique genes knocked out in all of data collected; -1 indicates all genes
    """
    def __init__(self,
                 b,
                 k_param,
                 k_sign_p,
                 hill,
                 decays,
                 noise_params,
                 cell_types,
                 noise_type='dpd',
                 sampling_state=15,
                 dt=0.01,
                 tech_noise_config=None,
                 add_outlier_effect=False,
                 add_lib_size_effect=False,
                 add_dropout_effect=False,
                 return_count_data=False,
                 # interventions
                 n_ko_genes=0):

        self.b = b
        self.k_param = k_param
        self.k_sign_p = k_sign_p
        self.hill = hill
        self.decays = decays
        self.noise_params = noise_params
        self.cell_types = cell_types
        self.noise_type = noise_type
        self.sampling_state = sampling_state
        self.dt = dt
        self.tech_noise_config = tech_noise_config
        self.add_outlier_effect = add_outlier_effect
        self.add_lib_size_effect = add_lib_size_effect
        self.add_dropout_effect = add_dropout_effect
        self.return_count_data = return_count_data
        self.n_ko_genes = n_ko_genes

    def __call__(self, rng, g, n_observations_obs, n_observations_int):

        # sample interaction terms K
        n_vars = g.shape[-1]
        k = onp.abs(self.k_param(rng, shape=(n_vars, n_vars)))
        effect_sgn = rng.binomial(1, self.k_sign_p(rng, shape=(n_vars, 1)), size=g.shape) * 2.0 - 1.0

        k = k * effect_sgn.astype(onp.float32)
        assert onp.array_equal(k != 0, effect_sgn != 0)
        assert set(onp.unique(effect_sgn)).issubset({-1.0, 0.0, 1.0})

        # sample number of cell types to be sampled (i.e. number of unique master regulator reproduction rates)
        n_cell_types = int(self.cell_types(rng, shape=1))
        assert n_cell_types > 0, "Need at least 1 cell type to be simulated"

        # master regulator basal reproduction rate
        basal_rates = self.b(rng, shape=(n_vars, n_cell_types)) # assuming 1 cell type is simulated

        # hill coeff
        hills = self.hill * onp.ones((n_vars, n_vars))

        # sample technical noise configuration
        if self.add_outlier_effect or self.add_lib_size_effect or self.add_dropout_effect:
            config = load_yaml(path=SERGIO_NOISE_CONFIG)
            assert self.tech_noise_config in config.keys(), f"tech_noise_config `{self.tech_noise_config}` " \
                                                            f"not in config keys: `{list(config.keys())}`"
            outlier_prob = rng.choice(config[self.tech_noise_config]["outlier_prob"])
            outlier_mean = rng.choice(config[self.tech_noise_config]["outlier_mean"])
            outlier_scale = rng.choice(config[self.tech_noise_config]["outlier_scale"])
            lib_size_mean = rng.choice(config[self.tech_noise_config]["lib_size_mean"])
            lib_size_scale = rng.choice(config[self.tech_noise_config]["lib_size_scale"])
            dropout_shape = rng.choice(config[self.tech_noise_config]["dropout_shape"])
            dropout_percentile = rng.choice(config[self.tech_noise_config]["dropout_percentile"]).astype(int)

        # sample targets for experiments performed (wild-type, knockout)
        ko_targets = []

        simulate_observ_data = n_observations_obs > 0
        if simulate_observ_data:
            ko_targets += [None]

        simulate_interv_data = n_observations_int > 0
        if simulate_interv_data:
            assert self.n_ko_genes != 0, f"Need n_ko_genes != 0 to have interventional data for SERGIO"
            n_ko_genes = n_vars if self.n_ko_genes == -1 else self.n_ko_genes
            ko_targets += sorted(rng.choice(n_vars, size=min(n_vars, n_ko_genes), replace=False).tolist())


        # simulate for wild type and each ko target
        data = defaultdict(lambda: defaultdict(list))
        for ko_target in ko_targets:

            if ko_target is None:
                # observational/wild type
                data_type = "obs"
                kout = onp.zeros(n_vars).astype(bool)
                number_sc = math.ceil(n_observations_obs / n_cell_types)

            else:
                # interventional/knockout
                data_type = "int"
                kout = onp.eye(n_vars)[ko_target].astype(bool)
                number_sc = math.ceil(n_observations_int / (n_cell_types * n_ko_genes))

            # setup simulator
            sim = SergioSimulator(
                rng=rng,
                number_genes=n_vars,
                number_bins=n_cell_types,
                number_sc=number_sc,
                noise_params=self.noise_params,
                noise_type=self.noise_type,
                decays=self.decays,
                sampling_state=self.sampling_state,
                kout=kout,
                dt=self.dt,
                safety_steps=10,
            )

            sim.custom_graph(
                g=g,
                k=k,
                b=basal_rates,
                hill=hills,
            )

            # run steady-state simulations
            assert number_sc >= 1, f"Need to have number_sc >= 1: number_sc {number_sc} data_type {data_type}"
            sim.simulate()

            # Get the clean simulated expression matrix after steady_state simulations
            # shape: [number_bins (#cell types), number_genes, number_sc (#cells per type)]
            expr = sim.getExpressions()

            # add technical noise (assume 1 experimental setting for wild type and KO experiments)
            # 1) outlier genes
            if self.add_outlier_effect:
                expr = sim.outlier_effect(expr, outlier_prob, outlier_mean, outlier_scale)

            # 2) library size
            if self.add_lib_size_effect:
                _, expr = sim.lib_size_effect(expr, lib_size_mean, lib_size_scale)

            # 3) dropout
            if self.add_dropout_effect:
                binary_ind = sim.dropout_indicator(expr, dropout_shape, dropout_percentile)
                expr = onp.multiply(binary_ind, expr)

            # 4) mRNA count data
            if self.return_count_data:
                expr = sim.convert_to_UMIcounts(expr)

            # Aggregate by concatenating gene expressions of all cell types into 2d array
            # [number_genes (#genes), number_bins * number_sc (#cells  = #cell_types * #cells_per_type)]
            expr_agg = onp.concatenate(expr, axis=1)

            # Now each row represents a gene and each column represents a simulated single-cell
            # Gene IDs match their row in this expression matrix
            # [number_bins * number_sc, number_genes]
            x = expr_agg.T
            x = rng.permutation(x, axis=0)

            # generate intervention mask
            # [number_bins * number_sc, number_genes] with True/False depending on whether gene was knocked out
            ko_mask = onp.tile(kout, (x.shape[0], 1)).astype(onp.float32)

            # advance rng outside for faithfullness/freshness of data in for loop
            rng = copy.deepcopy(sim.rng)

            data[data_type]["x"].append(x)
            data[data_type]["ko_mask"].append(ko_mask)


        # concatenate interventional data by interweaving rows to have balanced intervention target counts
        if simulate_observ_data:
            x_obs = onp.stack(data["obs"]["x"]).reshape(-1, n_vars, order="F")
            x_obs_msk = onp.stack(data["obs"]["ko_mask"]).reshape(-1, n_vars, order="F")
        else:
            x_obs = onp.zeros((0, n_vars))  # dummy
            x_obs_msk = onp.zeros((0, n_vars))  # dummy

        if simulate_interv_data:
            x_int = onp.stack(data["int"]["x"]).reshape(-1, n_vars, order="F")
            x_int_msk = onp.stack(data["int"]["ko_mask"]).reshape(-1, n_vars, order="F")
        else:
            x_int = onp.zeros((0, n_vars)) # dummy
            x_int_msk = onp.zeros((0, n_vars)) # dummy

        # clip number of observations to be invariant to n_cell_types due to rounding
        # [n_observations, n_vars, 2]
        x_obs = onp.stack([x_obs, x_obs_msk], axis=-1)[:n_observations_obs, :, :]
        x_int = onp.stack([x_int, x_int_msk], axis=-1)[:n_observations_int, :, :]

        assert x_obs.size != 0 or x_int.size != 0, f"Need to sample at least some observations; " \
                                                   f"got shapes x_obs {x_obs.shape} x_int {x_int.shape}"

        # collect data
        return Data(
            x_obs=x_obs,
            x_int=x_int,
            is_count_data=True,
        )


import functools
import numpy as onp

from avici.synthetic.noise_scale import init_noise_dist
from avici.synthetic.utils import draw_rff_params, sample_recursive_scm
from avici.synthetic import MechanismModel


class RFFAdditive(MechanismModel):
    """
    Random fourier feature functions with additive noise

    As n_rff goes to infinity, function returned corresponds to a sample from GP prior with RBF kernel
        k(x, y) = c^2 * exp(- |x-y|^2 / (2 * ls^2)

    Rahimi and Recht, 2007
    https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf

    Args:
        length_scale (Distribution): distribution for sampling kernel length scale used to sample a GP function.
            Example: `avici.synthetic.Uniform`
        output_scale (Distribution): distribution for sampling kernel output scale used to sample a GP function.
            Example: `avici.synthetic.Uniform`
        bias (Distribution): distribution for sampling GP function biases. Example: `avici.synthetic.Uniform`
        noise (Distribution): distribution for sampling additive noise. Example: `avici.synthetic.Gaussian`
        noise_scale (Distribution): if specified, distribution used to sample noise scale
            for `avici.synthetic.SimpleNoise`. Example: `avici.synthetic.Uniform`.
            Only one of `noise_scale` and `noise_scale_heteroscedastic` must be specified.
        noise_scale_heteroscedastic (Distribution): if specified, kwargs for
            initalizing `avici.synthetic.HeteroscedasticRFFNoise`.
            Only one of `noise_scale` and `noise_scale_heteroscedastic` must be specified.
        n_interv_vars (int): no. unique variables intervened upoin in all of data collected;
            -1 indicates all nodes
        interv_dist (Distribution): distribution for sampling intervention values.
            Example: `avici.synthetic.SignedUniform`

    """
    def __init__(self,
                length_scale,
                output_scale,
                bias,
                noise,
                noise_scale=None,
                noise_scale_heteroscedastic=None,
                n_interv_vars=0,
                interv_dist=None):

       assert interv_dist is not None or n_interv_vars == 0

       self.length_scale = length_scale
       self.output_scale = output_scale
       self.bias = bias
       self.noise = noise
       self.noise_scale = noise_scale
       self.noise_scale_heteroscedastic = noise_scale_heteroscedastic
       self.n_interv_vars = n_interv_vars
       self.interv_dist = interv_dist

    @staticmethod
    def _rff_mechanism(*, x, z, is_parent, bias, omega, b, w, c, n_rff):
        # [..., n_parents]
        x_parents = x[:, onp.where(is_parent)[0]]

        # feature map phi = cos(omega @ x + b)
        # [..., n_parents, n_rff], [..., n_parents] -> [..., n_rff]
        phi = onp.cos(onp.einsum('db,...d->...b', omega, x_parents) + b)

        # f(x) = w @ phi(x)
        # [..., n_rff], [..., n_rff] -> [...]
        f_j = onp.sqrt(2.0) * c * onp.einsum('b,...b->...', w, phi) / onp.sqrt(n_rff)

        # additive noise
        x_j = f_j + bias + z
        return x_j


    def __call__(self, rng, g, n_observations_obs, n_observations_int):

        # construct mechanism for each node
        n_vars = g.shape[-1]
        f = []
        for j in range(n_vars):
            # draw rff mechanism parameters for node j
            n_parents = int(g[:, j].sum().item())
            f_params = draw_rff_params(rng=rng, d=n_parents, length_scale=self.length_scale,
                                       output_scale=self.output_scale, n_rff=100)
            b = self.bias(rng, shape=(1,))

            # bind these parameters to mechanism function
            f.append(functools.partial(RFFAdditive._rff_mechanism, **f_params, bias=b))

        # construct noise distribution for each node
        nse = []
        for j in range(n_vars):
            # sample parameters and bind to sampling function
            # each call may sample a random noise scale, so call once per node
            n_parents = int(g[:, j].sum().item())
            nse.append(init_noise_dist(rng=rng,
                                       dim=n_parents,
                                       dist=self.noise,
                                       noise_scale=self.noise_scale,
                                       noise_scale_heteroscedastic=self.noise_scale_heteroscedastic))

        # sample recursively over g given functionals and endogenous noise distribution
        data = sample_recursive_scm(
            rng=rng,
            n_observations_obs=n_observations_obs,
            n_observations_int=n_observations_int,
            g=g,
            f=f,
            nse=nse,
            interv_dist=self.interv_dist,
            n_interv_vars=self.n_interv_vars,
        )
        return data


from pathlib import Path
import numpy as onp

from avici.definitions import PROJECT_DIR
from avici.utils.parse import load_data_config
from avici.buffer import Sampler
from avici.utils.data import onp_standardize_data


def simulate_data( d, n, *, n_interv=0, seed=0, domain=None, path=None, module_paths=None):
    """
    Helper function for simulating data from a pre-specified `domain` or a YAML domain configuration file at `path`.

    Args:
        d (int): number of variables in the system to be simulated
        n (int): number of observational data points to be sampled
        n_interv (int): number of interventional data points to be sampled
        seed (int): random seed
        domain (str): specifier of domain to be simulated. Currently implemented options:
            `lin-gauss`,
            `lin-gauss-heterosked`,
            `lin-laplace-cauchy`,
            `rff-gauss`,
            `rff-gauss-heterosked`,
            `rff-laplace-cauchy`,
            `gene-ecoli`
            (all `.yaml` file names inside `avici.config.examples`).
            Only one of `domain` and `path` must be specified.
        path (str): path to YAML domain configuration, like the examples in `avici.config`
        module_paths (str): path (or list of paths) to additional modules used in the domain configuration file

    Returns:
        tuple: the function returns a 3-tuple of
            - g (ndarray): `[d, d]` causal graph of `d` variables
            - x (ndarray): `[n + n_interv, d]` data matrix containing `n + n_interv` observations of the `d` variables
            - interv (ndarray): `[n + n_interv, d]` binary matrix indicating which nodes were intervened upon

    """
    rng = onp.random.default_rng(onp.random.SeedSequence(entropy=seed))

    # load example domain specification
    if domain is not None:
        assert path is None, "Only specify one of `domain` and `path`"
        abspath = PROJECT_DIR / f"avici/config/examples/{domain}.yaml"
    elif path is not None:
        assert domain is None, "Only specify one of `domain` and `path`"
        abspath = Path(path)
    else:
        raise KeyError("Specify either an an `avici.config.examples` domain (YAML name) or a path to a YAML config.")

    if abspath.is_file():
        kwargs = dict(n_observations_obs=n, n_observations_int=n_interv)
        spec_tree = load_data_config(abspath, force_kwargs=kwargs, abspath=True,
                                     module_paths=module_paths, load_modules=True)["data"]
        spec = spec_tree[next(iter(spec_tree))]
    else:
        raise KeyError(f"`{abspath}` does not exist.")

    # sample and concatenate all data
    data = Sampler.generate_data(
        rng,
        n_vars=d,
        spec_list=spec,
    )
    x = onp.concatenate([data["x_obs"], data["x_int"]], axis=-3)

    # standardize only if not real-valued data
    data = onp_standardize_data(data) if not data["is_count_data"] else data

    if n_interv:
        return data["g"].astype(int), x[..., 0], x[..., 1]
    else:
        return data["g"].astype(int), x[..., 0], None


import os
import json
import pickle
import functools
import jax
import jax.random as random
import jax.numpy as jnp
import haiku as hk
import optax
from optax._src.linear_algebra import global_norm
from jax.tree_util import tree_map

from deepdiff import DeepDiff
from typing import Any, NamedTuple
from pathlib import Path

import warnings
warnings.formatwarning = lambda msg, category, path, lineno, file: f"{path}:{lineno}: {category.__name__}: {msg}\n"

from avici.definitions import CHECKPOINT_KWARGS


def get_first(super_pytree):
    """Gets values from the first device."""
    return tree_map(lambda leaf: leaf[0], super_pytree)


def make_serializable(d):
    if isinstance(d, dict):
        return {k: make_serializable(v) for k, v in d.items()}
    elif isinstance(d, list):
        return [make_serializable(v) for v in d]
    elif callable(d):
        if hasattr(d, "__name__"):
            return d.__name__
        else:
            return type(d).__name__
    elif isinstance(d, Path):
        return str(d)
    else:
        return d


def fix_json_loading(d):
    if isinstance(d, dict):
        return {((int(k) if k.isdigit() else k) if isinstance(k, str) else k):
                    fix_json_loading(v) for k, v in d.items()}
    elif isinstance(d, list):
        return [fix_json_loading(v) for v in d]
    elif isinstance(d, Path):
        return str(d)
    else:
        return d


class ModelState(NamedTuple):
    step: int
    rng: Any
    opt_state: Any
    params: Any
    dual: Any
    dual_penalty_polyak: Any
    ave_params: Any  # polyak average


class Updater:
    """
    A stateless abstraction around an init_fn/update_fn pair.
    Creates a `ModelState` object and updates it.
    """

    def __init__(self, *, net_init, loss_fn, opt,
                 acyclicity_dual_lr,
                 acyclicity_inner_step,
                 acyclicity_burnin,
                 acyclicity_warmup,
                 polyak_rate=1e-4):

        self._net_init = net_init
        self._loss_fn = loss_fn
        self._opt = opt
        self.distributed = False
        self.acyclicity_dual_lr = acyclicity_dual_lr
        self.acyclicity_inner_step = acyclicity_inner_step
        self.acyclicity_burnin = acyclicity_burnin
        self.acyclicity_warmup = acyclicity_warmup
        self.polyak_rate = polyak_rate

    @functools.partial(jax.jit, static_argnums=0)
    def init(self, rng, x):
        """Initializes model state of the updater."""
        out_rng, init_rng = random.split(rng)
        params = self._net_init(init_rng, x, True) # is_training = True
        opt_state = self._opt.init(params)
        return ModelState(
            step=0,
            rng=out_rng,
            opt_state=opt_state,
            params=params,
            dual=jnp.array(0.0),
            dual_penalty_polyak=jnp.array(0.0),
            ave_params=params,
        )

    @functools.partial(jax.jit, static_argnums=0)
    def update(self, state, batch, t):
        """Updates the ModelState `state` using data `batch` and returns metrics."""

        new_rng, step_rng = random.split(state.rng)

        # compute loss and gradient
        (loss, aux), loss_grads = jax.value_and_grad(self._loss_fn, has_aux=True)(state.params, state.dual, step_rng, batch, t, True)  # is_training = True

        # optimizer step on params
        opt_update, opt_state = self._opt.update(loss_grads, state.opt_state, state.params) # some optimizers need params
        params = optax.apply_updates(state.params, opt_update)

        # track polyak average of params
        ave_params = optax.incremental_update(
            params, state.ave_params, step_size=0.001)

        # dual step
        dual_penalty = jnp.maximum(aux["acyc"], jnp.array(0.0))
        dual_penalty_ave = jax.lax.cond(
            t == 0,
            lambda _: dual_penalty,
            lambda _: optax.incremental_update(dual_penalty, state.dual_penalty_polyak, step_size=self.polyak_rate),
            operand=None)

        if self.acyclicity_warmup:
            dual_lr = jnp.minimum(self.acyclicity_dual_lr, t * self.acyclicity_dual_lr / self.acyclicity_burnin)
            effective_burnin = 0
        else:
            dual_lr = self.acyclicity_dual_lr
            effective_burnin = self.acyclicity_burnin

        dual = jax.lax.cond(
            (jnp.mod(t, self.acyclicity_inner_step) == 0) & (t > effective_burnin),
            lambda _: state.dual + dual_lr * dual_penalty_ave,
            lambda _: state.dual,
            operand=None)

        # state
        new_state = ModelState(
            step=state.step + 1,
            rng=new_rng,
            opt_state=opt_state,
            params=params,
            dual=dual,
            dual_penalty_polyak=dual_penalty_ave,
            ave_params=ave_params,
        )
        # log scalars
        metrics = {
            "loss": loss,
            "dual": dual,
            "grad_norm": global_norm(loss_grads),
            **aux,
        }
        return new_state, metrics


class SuperUpdater:
    """
    `Updater` with distributed training (pmap) functionality.
    """

    def __init__(self, *, net_init, loss_fn, opt,
                 acyclicity_dual_lr,
                 acyclicity_inner_step,
                 acyclicity_burnin,
                 acyclicity_warmup,
                 local_device_count,
                 polyak_rate=1e-4):
        self._net_init = net_init
        self._loss_fn = loss_fn
        self._opt = opt
        self.distributed = True
        self.local_device_count = local_device_count
        self.acyclicity_dual_lr = acyclicity_dual_lr
        self.acyclicity_inner_step = acyclicity_inner_step
        self.acyclicity_burnin = acyclicity_burnin
        self.acyclicity_warmup = acyclicity_warmup
        self.polyak_rate = polyak_rate

    @functools.partial(jax.pmap, static_broadcasted_argnums=0)
    def init(self, rng, x):
        """Initializes model state of the updater."""
        out_rng, init_rng = random.split(rng)
        params = self._net_init(init_rng, x, True)  # is_training = True
        opt_state = self._opt.init(params)
        return ModelState(
            step=0,
            rng=out_rng,
            opt_state=opt_state,
            params=params,
            dual=jnp.array(0.0),
            dual_penalty_polyak=jnp.array(0.0),
            ave_params=params,
        )

    # @jax.jit # comment this in, on top of pmap, to debug NaN's inside pmap
    @functools.partial(jax.pmap, in_axes=(0, 0, 0, None), axis_name="i", static_broadcasted_argnums=0, donate_argnums=(1, 2))
    def update(self, state, batch, t):
        """Updates the ModelState `state` using data `batch` and returns metrics."""

        new_rng, step_rng = random.split(state.rng)

        # compute loss and gradient
        (loss, aux), loss_grads = jax.value_and_grad(self._loss_fn, has_aux=True)(state.params, state.dual, step_rng, batch, t, True)  # is_training = True

        # take the mean of the gradients across all data-parallel replicas
        loss_grads = jax.lax.pmean(loss_grads, axis_name="i")

        # optimizer step on params (will perform same step on every device due to above line)
        opt_update, opt_state = self._opt.update(loss_grads, state.opt_state, state.params) # some optimizers need params
        params = optax.apply_updates(state.params, opt_update)

        # track polyak average of params
        ave_params = optax.incremental_update(
            params, state.ave_params, step_size=0.001)

        # dual step
        dual_penalty = jnp.maximum(jax.lax.pmean(aux["acyc"], axis_name="i"), jnp.array(0.0))
        dual_penalty_ave = jax.lax.cond(
            t == 0,
            lambda _: dual_penalty,
            lambda _: optax.incremental_update(dual_penalty, state.dual_penalty_polyak, step_size=self.polyak_rate),
            operand=None)

        if self.acyclicity_warmup:
            dual_lr = jnp.minimum(self.acyclicity_dual_lr, t * self.acyclicity_dual_lr / self.acyclicity_burnin)
            effective_burnin = 0
        else:
            dual_lr = self.acyclicity_dual_lr
            effective_burnin = self.acyclicity_burnin

        dual = jax.lax.cond(
            (jnp.mod(t, self.acyclicity_inner_step) == 0) & (t > effective_burnin),
            lambda _: state.dual + dual_lr * dual_penalty_ave,
            lambda _: state.dual,
            operand=None)

        # state
        new_state = ModelState(
            step=state.step + 1,
            rng=new_rng,
            opt_state=opt_state,
            params=params,
            dual=dual,
            dual_penalty_polyak=dual_penalty_ave,
            ave_params=ave_params,
        )

        # log scalars
        # note: we log the mean across all hosts/devices
        metrics = {
            "loss": loss,
            "dual": dual,
            "grad_norm": global_norm(loss_grads),
            **aux,
        }
        metrics = jax.lax.pmean(metrics, axis_name="i")

        return new_state, metrics



class CheckpointingUpdater:
    """A didactic checkpointing wrapper around an `Updater` or `SuperUpdater`."""

    def __init__(self, inner, checkpoint_dir, checkpoint_every_n=10000, base_str="checkpoint_", save_kwargs=None):
        self._inner = inner
        self._checkpoint_dir = checkpoint_dir
        self._checkpoint_every_n = checkpoint_every_n
        self._base_str = base_str
        self._save_kwargs = save_kwargs
        self._just_loaded = False # avoids serialization immediately after loading

    def _checkpoint_paths(self):
        return sorted([p for p in os.listdir(self._checkpoint_dir) if self._base_str in p])

    def checkpoint(self, state, step):
        path = os.path.join(self._checkpoint_dir, f'{self._base_str}{step:07d}.pkl')
        if self._inner.distributed:
            # For distributed training, only save 1 copy of model state
            state = tree_map(lambda leaf: leaf[0], state)
        checkpoint_state = jax.device_get(state)
        print(f'Serializing experiment state to {path}', flush=True)
        with open(path, 'wb') as f:
            pickle.dump(checkpoint_state, f)
        return

    def init(self, rng, data):
        """Initialize experiment state. """
        if not os.path.exists(self._checkpoint_dir) or not self._checkpoint_paths():
            os.makedirs(self._checkpoint_dir, exist_ok=True)
            print(f'Checkpoint directory at {self._checkpoint_dir}', flush=True)
            if self._save_kwargs is not None:
                with open(os.path.join(self._checkpoint_dir, CHECKPOINT_KWARGS), "w") as file:
                    json.dump(make_serializable(self._save_kwargs), file, indent=4, sort_keys=True)
            return self._inner.init(rng, data)
        else:
            last_checkpoint = os.path.join(self._checkpoint_dir, self._checkpoint_paths()[-1])
            print(f'Loading latest checkpoint from {last_checkpoint}', flush=True)
            if self._save_kwargs is not None:
                with open(os.path.join(self._checkpoint_dir, CHECKPOINT_KWARGS), "r") as file:
                    loaded_kwargs = json.load(file)
                    loaded_kwargs = fix_json_loading(loaded_kwargs)
                    if loaded_kwargs != make_serializable(self._save_kwargs):
                        diff = DeepDiff(loaded_kwargs, make_serializable(self._save_kwargs)).pretty()
                        warn_str = f"Specified save_kwargs and those found in checkpoint directory don't match: \n"
                        warnings.warn(warn_str + diff)

            with open(last_checkpoint, 'rb') as f:
                state = pickle.load(f)
                if self._inner.distributed:
                    # For distributed training, replicate copy of model state on each device
                    devices = jax.local_devices()
                    state = tree_map(lambda leaf: jax.device_put_sharded(len(devices) * [leaf], devices), state)
                self._just_loaded = True
                return state

    def update(self, state, data, step):
        """Update experiment state. """
        # `step` is maintained separately to allow for JAX async dispatch inside `state`
        # https://jax.readthedocs.io/en/latest/async_dispatch.html
        if (step % self._checkpoint_every_n) == 0 and step != 0:
            if self._just_loaded:
                self._just_loaded = False
            else:
                self.checkpoint(state, step)

        state, out = self._inner.update(state, data, step)
        return state, out



import zipfile
import functools
import inspect
import shutil
import warnings
from tqdm.auto import tqdm
from pathlib import Path

import jax
import numpy as onp
import jax.numpy as jnp

from avici.model import BaseModel, InferenceModel
from avici.utils.load import load_checkpoint
from avici.utils.data_jax import jax_standardize_default_simple, jax_standardize_count_simple
from avici.definitions import CACHE_SUBDIR, CHECKPOINT_KWARGS, HUGGINGFACE_REPO, \
    MODEL_NEURIPS_LINEAR, MODEL_NEURIPS_RFF, MODEL_NEURIPS_GRN, MODEL_NEURIPS_SCM_V0

from huggingface_hub import hf_hub_download


class AVICIModel:
    """
    Wrapper class for `avici.model.InferenceModel` that allows downloading and storing parameters
    and making predictions while exposing less functionality of the inference model.

    Args:
        params (pytree): parameters for the infernece model, extracted from `avici.backbone.ModelState.params`
        model (avici.model.InferenceModel`)
        expects_counts (bool): whether or not model expects counts. Since we pass the standardizer manually,
            this is technically not needed, but it is used to issue a warning in case the model expects counts
            but gets non-integer data (in case, for example, the data is accidentally standardized before `__call__`.
        standardizer: function standardizing the data prior to the forward pass.
            Default is `jax_standardize_default_simple` from `avici.utils.data_jax`, which performs the usual
            z-standardization for real-valued data.

    """

    def __init__(self, *,
                 params,
                 model,
                 expects_counts,
                 standardizer=jax_standardize_default_simple,
                 ):

        self.params = params
        self.expects_counts = expects_counts
        self._model = model
        self._standardizer = standardizer


    @functools.partial(jax.jit, static_argnums=(0,))
    def __call_main__(self, x, interv=None):
        # concatenate intervention indicators
        if interv is None:
            x = jnp.stack([x, jnp.zeros_like(x)], axis=-1)
        else:
            assert x.shape == interv.shape
            x = jnp.stack([x, interv.astype(x.dtype)], axis=-1)

        # standardize data
        x = self._standardizer(x)

        # forward pass through inference model
        g_edges_prob = self._model.infer_edge_probs(self.params, x)
        return g_edges_prob


    def __call__(self, x, interv=None, return_probs=True):
        """
        Wraps __call_main__ to do some tests and warnings before jax.jit

        Args:
            x: `[n, d]` Real-valued data matrix
            interv(optional): Binary matrix of the same shape as `x`
                with `interv[i,j] == 1` iff node `j` was intervened upon in observation `i`
            return_probs(optional): Whether to return probability estimates for each edge. Defaults to `True` as
                the computational cost is the same. `False` simply clips the predictions to 0 and 1 using a decision
                threshold of 0.5. Other thresholds achieve a different true positive vs
                false positive trade-off and can be computed using the probabilities returned with `True`.

        Returns:
            `[d, d]` adjacency matrix of predicted edge probabilities
        """
        x_type = type(x)

        # check that interv mask is binary
        if interv is not None:
            assert x.shape == interv.shape, "`x` and `interv` must have the same shape when provided."
            interv_is_binary = onp.all(onp.isclose(interv, 0) | onp.isclose(interv, 1))
            assert interv_is_binary, "Intervention mask `interv` has to be binary."
            interv = onp.around(interv, 0, out=interv)

        # check that x contains integers if model expects counts
        if self.expects_counts:
            x_is_int = onp.allclose(onp.mod(x, 1), 0)
            if not x_is_int:
                warnings.warn(f"Model expects count data but `x` contains non-integer values. ")

        # main call
        out = self.__call_main__(x=x, interv=interv)

        # if desired, threshold output
        if not return_probs:
            out = (out > 0.5).astype(int)

        # convert to input type
        if x_type == onp.ndarray:
            out = onp.array(out)

        return out


def load_pretrained(download=None, force_download=False, checkpoint_dir=None, cache_path="./", expects_counts=None,
                    verbose=True):
    f"""
    Loads a pretrained AVICI model

    Args:
        download (str, optional): Specifier for existing pretrained model checkpoint to be downloaded (online)
        force_download (bool, optional): Whether to force re-download of model checkpoint specified via `download`
        checkpoint_dir (str, optional): Path to *folder* containing both the model checkpoint `<name>.pkl` and
            the model kwargs `{CHECKPOINT_KWARGS}` files (stored locally).
        cache_path (str, optional): Path used as cache directory for storing the
            downloaded model checkpoints (Default: `./cache`)
        expects_counts (bool, optional): Whether model expects count data, for data standardization purposes.
            Required when providing `checkpoint_dir`.
        verbose (bool, optional): Whether to print path information

    Returns:
        `avici.AVICIModel`
    """
    assert (download is None and checkpoint_dir is not None) \
        or (download is not None and checkpoint_dir is None), "Specify exactly one of `download` or `checkpoint_dir`"

    # get figshare source link and select standardization function (bind heldout_data as `None`)
    if download is not None:
        if download == "neurips-linear":
            paths = MODEL_NEURIPS_LINEAR
            expects_counts = False
        elif download == "neurips-rff":
            paths = MODEL_NEURIPS_RFF
            expects_counts = False
        elif download == "neurips-grn":
            paths = MODEL_NEURIPS_GRN
            expects_counts = True
        elif download == "scm-v0":
            paths = MODEL_NEURIPS_SCM_V0
            expects_counts = False
        else:
            raise ValueError(f"Unknown download specified: `{download}`")

        # download source if not yet downloaded
        if cache_path == "./":
            try:
                cache_path = Path(cache_path).resolve() / CACHE_SUBDIR
            except FileNotFoundError:
                raise FileNotFoundError("Could not resolve default cache_path `./` for downloads. "
                                        "Please specify the download location by specifying `cache_path`")
            if verbose:
                print(f"Using default cache_path: `{cache_path}`")

        # download from huggingface
        model_paths = []
        for path in paths:
            subfolder, filename = Path(path).parent, Path(path).name
            file_path = hf_hub_download(
                repo_id=HUGGINGFACE_REPO,
                subfolder=subfolder,
                filename=filename,
                cache_dir=cache_path,
                force_download=force_download,
            )
            model_paths.append(Path(file_path))

        assert all([p.parent == model_paths[0].parent for p in model_paths]), \
            f"Folder structure error. All files should be from the same model folder. "

        root_path = model_paths[0].parent

    else:
        assert expects_counts is not None, "When loading custom model checkpoint, need to specify `expects_counts` as "\
                                           "`True`/`False` to ensure proper data standardization. Specifying " \
                                           "`False` should be the default and implies the data is z-standardized."
        root_path = Path(checkpoint_dir)
        assert root_path.exists(), f"The provided checkpoint_dir `{checkpoint_dir}` does not exist. " \
                                    f"In case you provided a relative path, try the absolute path instead."
        assert root_path.is_dir(), "The provided checkpoint_dir is not a directory. You should specify the path to " \
                                    f"a folder that contains the model `.pkl` (and `{CHECKPOINT_KWARGS}`), " \
                                    "not the path to the `.pkl` alone."
        assert (root_path.parent / CHECKPOINT_KWARGS).exists(), f"The provided checkpoint_dir `{checkpoint_dir}` " \
                                    f"does not contain the file `{CHECKPOINT_KWARGS}`, which is required. "
        print(f"Loading local checkpoint from `{root_path}`")

    # select data standardizer
    if expects_counts:
        standardizer = jax_standardize_count_simple
    else:
        standardizer = jax_standardize_default_simple

    # load checkpoint
    state, loaded_config = load_checkpoint(root_path)
    inference_model_kwargs = loaded_config["inference_model_kwargs"]
    try:
        neural_net_kwargs = loaded_config["neural_net_kwargs"]["model_kwargs"] # legacy compatibility
    except KeyError:
        neural_net_kwargs = loaded_config["neural_net_kwargs"]

    # discard deprecated kwargs from checkpoint
    sig = inspect.signature(InferenceModel).parameters
    deprec = list(filter(lambda key: key not in sig, inference_model_kwargs.keys()))
    for k in deprec:
        del inference_model_kwargs[k]
        # print(f"Ignoring deprecated kwarg `{k}` loaded from `inference_model_kwargs` in checkpoint")

    # init model
    inference_model = InferenceModel(**inference_model_kwargs,
                                     model_class=BaseModel,
                                     model_kwargs=neural_net_kwargs)

    model = AVICIModel(
        params=state.params,
        model=inference_model,
        expects_counts=expects_counts,
        standardizer=standardizer,
    )

    return model







import numpy as onp
from sklearn import metrics as sklearn_metrics


def shd(g, h):
    """
    Computes pairwise Structural Hamming distance, i.e.
    the number of edge insertions, deletions or flips in order to transform one graph to another
        - this means, edge reversals do not double count
        - this means, getting an undirected edge wrong only counts 1

    Args:
        g:  [..., d, d]
        h:  [..., d, d]
    """
    assert g.ndim == h.ndim
    abs_diff =  onp.abs(g - h)
    mistakes = abs_diff + onp.swapaxes(abs_diff, -2, -1)  # mat + mat.T (transpose of last two dims)

    # ignore double edges
    mistakes_adj = onp.where(mistakes > 1, 1, mistakes)

    return onp.triu(mistakes_adj).sum((-1, -2))


def n_edges(g):
    """
    Args:
        g:  [..., d, d]
    """
    return g.sum((-1, -2))


def is_acyclic(g):
    """
       Args:
           g:  [d, d]
       """
    n_vars = g.shape[-1]
    mat = onp.eye(n_vars) + g / n_vars
    mat_pow = onp.linalg.matrix_power(mat, n_vars)
    acyclic_constr = onp.trace(mat_pow) - n_vars
    return onp.isclose(acyclic_constr, 0.0)


def is_cyclic(g):
    """
    Args:
        g:  [d, d]
    """
    return not is_acyclic(g)


def classification_metrics(true, pred):
    """
    Args:
        true:  [...]
        pred:  [...]
    """
    true_flat = true.reshape(-1)
    pred_flat = pred.reshape(-1)

    if onp.sum(pred_flat) > 0 and onp.sum(true_flat) > 0:
        precision, recall, f1, _ = sklearn_metrics.precision_recall_fscore_support(
            true_flat, pred_flat, average="binary")
        return {
            "precision": precision,
            "recall": recall,
            "f1": f1,
        }

    elif onp.sum(pred_flat) == 0 and onp.sum(true_flat) == 0:
        # no true positives, and no positives were predicted
        return {
            "precision": 1.0,
            "recall": 1.0,
            "f1": 1.0,
        }
    else:
        # no true positives, but we predicted some positives
        return {
            "precision": 0.0,
            "recall": 0.0,
            "f1": 0.0,
        }


def threshold_metrics(true, pred):
    """
    Args:
        true:  [...]
        pred:  [...]
    """
    true_flat = true.reshape(-1)
    pred_flat = pred.reshape(-1)

    if onp.sum(pred_flat) > 0 and onp.sum(true_flat) > 0:
        fpr, tpr, _ = sklearn_metrics.roc_curve(true_flat, pred_flat)
        precision, recall, _ = sklearn_metrics.precision_recall_curve(true_flat, pred_flat)
        ave_prec = sklearn_metrics.average_precision_score(true_flat, pred_flat)
        roc_auc = sklearn_metrics.auc(fpr, tpr)
        prc_auc = sklearn_metrics.auc(recall, precision)

        return {
            "auroc": roc_auc,
            "auprc": prc_auc,
            "ap": ave_prec,
        }

    elif onp.sum(pred_flat) == 0 and onp.sum(true_flat) == 0:
        # no true positives, and no positives were predicted
        return {
            "auroc": 1.0,
            "auprc": 1.0,
            "ap": 1.0,
        }

    else:
        # no true positives, but we predicted some positives
        return {
            "auroc": 0.5,
            "auprc": 0.0,
            "ap": 0.0,
        }


def _calibration_stats(*, y_pred, y_true, n_bins):
    assert y_true.ndim == 1 and y_pred.ndim == 1
    assert y_true.shape[0] == y_pred.shape[0]

    counts_per_bin = onp.zeros(n_bins)
    prob_pred = onp.zeros(n_bins)
    prob_true = onp.zeros(n_bins)

    bins = onp.linspace(0.0, 1.0 + 1e-8, n_bins + 1)
    for i in range(n_bins):
        lo, hi = bins[i], bins[i + 1]

        # get predicted probs that fall into this bin
        mask = (lo <= y_pred) & (y_pred < hi)
        y_pred_in_bin = y_pred[mask]
        n_preds_in_bin = y_pred_in_bin.shape[0]
        if n_preds_in_bin:
            # get true labels for these predicted probs
            y_true_in_bin = y_true[mask]
            y_true_in_bin_0 = y_true_in_bin == 0
            y_true_in_bin_1 = y_true_in_bin == 1
            assert y_true_in_bin_0.sum() + y_true_in_bin_1.sum() == y_true_in_bin.shape[0]

            # compute coverage
            prob_pred[i] = y_pred_in_bin.sum() / n_preds_in_bin
            prob_true[i] = y_true_in_bin_1.sum() / n_preds_in_bin

        else:
            prob_pred[i] = onp.nan
            prob_true[i] = onp.nan

        counts_per_bin[i] = n_preds_in_bin

    nan_mask = onp.isnan(prob_pred) | onp.isnan(prob_true)
    prob_pred = prob_pred[~nan_mask]
    prob_true = prob_true[~nan_mask]
    counts_per_bin = counts_per_bin[~nan_mask]

    assert prob_pred.shape == prob_true.shape
    assert prob_pred.shape == counts_per_bin.shape

    return prob_pred, prob_true, counts_per_bin, bins


def make_calibration_stats(true_inp, pred_inp, n_bins=10):

    # convert list into one flattened array
    if type(true_inp) == list and type(pred_inp) == list:
        assert len(true_inp) == len(pred_inp)
    else:
        true_inp = [true_inp]
        pred_inp = [pred_inp]

    y_true_all = onp.concatenate(true_inp).reshape(-1)
    y_pred_all = onp.concatenate(pred_inp).reshape(-1)

    # compute calibration line using full aggregate
    prob_pred_all, prob_true_all, counts_per_bin_all, bins_all = \
        _calibration_stats(y_pred=y_pred_all, y_true=y_true_all, n_bins=n_bins)

    # compute expected calibration error (ece) and brier score, batches of datasets to compute error
    eces, briers = [], []
    for y_true, y_pred in zip(true_inp, pred_inp):
        y_pred_single = y_pred.reshape(-1)
        y_true_single = y_true.reshape(-1)
        prob_pred_single, prob_true_single, counts_per_bin_single, _ = \
            _calibration_stats(y_pred=y_pred_single, y_true=y_true_single, n_bins=n_bins)

        # ece https://arxiv.org/pdf/1706.04599.pdf
        ece = onp.sum(onp.abs(prob_true_single - prob_pred_single) * counts_per_bin_single / y_pred_single.shape[0])
        eces.append(ece)

        # brier
        brier = onp.mean((y_pred_single - y_true_single) ** 2.0)
        briers.append(brier)

    return dict(
        prob_pred=prob_pred_all,
        prob_true=prob_true_all,
        counts_per_bin=counts_per_bin_all,
        y_pred=y_pred_all,
        bins=bins_all,
        eces=eces,
        briers=briers,
    )


import sys
# import importlib.util
import importlib
import itertools
import warnings
from collections import defaultdict
from pathlib import Path, PosixPath

import avici.synthetic
from avici.definitions import YAML_CLASS, YAML_MODULES, REAL_DATASET_KEY
from avici.synthetic import SyntheticSpec, CustomClassWrapper
from avici.utils.load import load_yaml


def cartesian_dict(d):
    """
    Cartesian product of nested dict of lists
    Example:

    d = {'s1': {'a': 0,
                'b': [0, 1, 2]},
         's2': {'c': [0, 1],
                'd': [0, 1]}}

    yields
        {'s1': {'a': 0, 'b': 0}, 's2': {'c': 0, 'd': 0}}
        {'s1': {'a': 0, 'b': 0}, 's2': {'c': 0, 'd': 1}}
        {'s1': {'a': 0, 'b': 0}, 's2': {'c': 1, 'd': 0}}
        {'s1': {'a': 0, 'b': 0}, 's2': {'c': 1, 'd': 1}}
        {'s1': {'a': 0, 'b': 1}, 's2': {'c': 0, 'd': 0}}
        ...
    """
    if type(d) in [dict, defaultdict]:
        keys, values = d.keys(), d.values()
        for c in itertools.product(*(cartesian_dict(v) for v in values)):
            yield dict(zip(keys, c))
    elif type(d) == list:
        for c in d:
            yield from cartesian_dict(c)
    else:
        yield d


def get_list_leaves(d):
    if type(d) in [dict, defaultdict]:
        opt = {}
        for k, v in d.items():
            if type(v) == list:
                opt[k] = v
        if opt:
            return cartesian_dict(opt)
        else:
            return []


def _parse_config_tree(config, force_kwargs=None, module_paths=None):

    if type(config) in [dict, defaultdict]:

        # check if `config` specifies a class
        is_class = YAML_CLASS in config.keys()

        # class
        if is_class:
            # recurse for kwargs
            name = config[YAML_CLASS]
            kwargs = {k: _parse_config_tree(subconfig, force_kwargs=force_kwargs, module_paths=module_paths)
                      for k, subconfig in config.items() if k != YAML_CLASS}

            # check for class in existing module
            try:
                cls = getattr(avici.synthetic, name)
                new = cls(**kwargs)

            # if class does not exist:
            except AttributeError:
                if module_paths is None:
                    raise SyntaxError(f"{YAML_CLASS} `{name}` of data config is not defined in existing "
                                      f"or registered modules. \nSpelled correctly and specified additional module paths?")
                # parse as custom class
                # if the class actually does not exist, loading will fail inside `load_custom_modules`
                new = CustomClassWrapper(name=name, kwargs=kwargs, paths=module_paths)

        # if not a class, recurse
        else:
            new = {key: _parse_config_tree(subconfig, force_kwargs=force_kwargs, module_paths=module_paths)
                   for (key, subconfig) in config.items()}

            # forcing key-value assignment if passed (e.g. n_observations)
            if force_kwargs is not None:
                for key in filter(lambda k: k in force_kwargs.keys(), new.keys()):
                    new[key] = force_kwargs[key]

    elif type(config) == list:
        # recurse
        new = [_parse_config_tree(subconfig, force_kwargs=force_kwargs, module_paths=module_paths)
               for subconfig in config]

    else:
        assert (type(config) in [int, float, bool, str, PosixPath] or config is None), f"Unknown yaml entry `{config}`"
        new = config

    return new


def init_custom_classes(config):

    # check if `config` specifies a `CustomModuleWrapper`
    if type(config) == CustomClassWrapper:

        # recurse on kwargs
        kwargs = init_custom_classes(config.kwargs)

        # load additional modules
        loaded_modules = []
        for p in config.paths:
            module_name = f"avici.synthetic._custom_.{p.stem}"
            module_spec = importlib.util.spec_from_file_location(module_name, p)
            sys.modules[module_name] = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(sys.modules[module_name])
            loaded_modules.append(sys.modules[module_name])

        # instantiate custom class
        new = None
        for mod in loaded_modules:
            try:
                cls = getattr(mod, config.name)
                new = cls(**kwargs)
                break
            except AttributeError:
                pass

        if new is None:
            raise SyntaxError(f"{YAML_CLASS} `{config.name}` of data config is not defined in existing "
                              f"or registered modules. \nSpelled correctly and specified additional module paths?")

    # else recurse
    elif type(config) == SyntheticSpec:
        new = SyntheticSpec(**init_custom_classes(config._asdict()))

    elif type(config) in [dict, defaultdict]:
        new = {key: init_custom_classes(subconfig) for (key, subconfig) in config.items()}

    elif type(config) == list:
        new = [init_custom_classes(subconfig) for subconfig in config]

    elif callable(config):
        new = config

    else:
        assert (type(config) in [int, float, bool, str, PosixPath] or config is None), f"Unknown entry `{config}`"
        new = config

    return new


def load_data_config(path, force_kwargs=None, abspath=False, module_paths=None, load_modules=False):
    """Load yaml config for data specification"""

    config = load_yaml(Path(path).resolve(strict=True), abspath=abspath)
    if config is None:
        raise SyntaxError("`config` is None; make sure there are no false tabs and indents in the .yaml file")

    spec = {"data": {}}

    # add meta info (all but "data" key)
    for key, val in config.items():
        if key not in spec:
            spec[key] = val

    # additional module paths
    additional_paths = []
    if YAML_MODULES in spec:
        ps = spec[YAML_MODULES] if isinstance(spec[YAML_MODULES], list) else [spec[YAML_MODULES]]
        additional_paths += list(map(lambda p: Path(p).resolve(strict=True), ps))

    if module_paths is not None:
        ps = module_paths if isinstance(module_paths, list) else [module_paths]
        additional_paths += list(map(lambda p: Path(p).resolve(strict=True), ps))

    additional_paths = list(set(additional_paths)) if additional_paths else None

    # real data case
    if REAL_DATASET_KEY in spec:
        return spec

    # process data field by expanding options lists with cartesian product
    spec["data"]["train"] = list(cartesian_dict(config["data"]))
    spec["data"]["val"] = list(cartesian_dict(config["data"]))

    for k in spec["data"].keys():
        if len(spec["data"][k]) >= 1000:
            warnings.warn(f"{k} config defines {len(spec['data'][k])} generative models")

    # init functions in spec
    spec = _parse_config_tree(spec, force_kwargs=force_kwargs, module_paths=additional_paths)

    # load modules; need to do this inside multiprocessing workers, otherwise pickle breaks
    if load_modules:
        spec = init_custom_classes(spec)

    # instantiate Synthetic data objects
    for descr, config_testcases in spec["data"].items():
        spec["data"][descr] = [SyntheticSpec(**arg) for arg in config_testcases]

    return spec



import sys
import numpy as np

class Gene(object):

    def __init__(self, geneID, geneType, binID = -1):

        """
        geneType: 'MR' master regulator or 'T' target
        bindID is optional
        """

        self.ID = geneID
        self.Type = geneType
        self.binID = binID
        self.Conc = []
        self.Conc_S = []
        self.dConc = []
        self.k = [] #For dynamics simulation it stores k1 to k4 for Rung-Kutta method, list of size 4 * num_c_to_evolve
        self.k_S = [] #For dynamics simulation it stores k1 to k4 for Rung-Kutta method, list of size 4 * num_c_to_evolve
        self.simulatedSteps_ = 0
        self.converged_ = False
        self.converged_S_ = False
        self.ss_U_ = 0.0 #This is the steady state concentration of Unspliced mRNA
        self.ss_S_ = 0.0 #This is the steady state concentration of Spliced mRNA

    def append_Conc (self, currConc):
        if isinstance(currConc, list):
            if currConc[0] < 0.0:
                self.Conc.append([0.0])
            else:
                self.Conc.append(currConc)
        else:
            if currConc < 0.0:
                self.Conc.append(0.0)
            else:
                self.Conc.append(currConc)


    def append_Conc_S (self, currConc):
        if isinstance(currConc, list):
            if currConc[0] < 0.0:
                self.Conc_S.append([0.0])
            else:
                self.Conc_S.append(currConc)
        else:
            if currConc < 0.0:
                self.Conc_S.append(0.0)
            else:
                self.Conc_S.append(currConc)

    def append_dConc (self, currdConc):
        self.dConc.append(currdConc)

    def append_k (self, list_currK):
        self.k.append(list_currK)

    def append_k_S (self, list_currK):
        self.k_S.append(list_currK)

    def del_lastK_Conc(self, K):
        for k in range(K):
            self.Conc.pop(-1)

    def del_lastK_Conc_S(self, K):
        for k in range(K):
            self.Conc_S.pop(-1)

    def clear_Conc (self):
        """
        This method clears all the concentrations except the last one that may
        serve as intial condition for rest of the simulations
        """
        self.Conc = self.Conc[-1:]

    def clear_dConc (self):
        self.dConc = []

    def incrementStep (self):
        self.simulatedSteps_ += 1

    def setConverged (self):
        self.converged_ = True

    def setConverged_S (self):
        self.converged_S_ = True

    def set_scExpression(self, list_indices):
        """
        selects input indices from self.Conc and form sc Expression
        """
        self.scExpression = np.array(self.Conc)[list_indices]

    def set_ss_conc_U(self, u_ss):
        if u_ss < 0.0:
            u_ss = 0.0

        self.ss_U_ = u_ss

    def set_ss_conc_S(self, s_ss):
        if s_ss < 0.0:
            s_ss = 0.0

        self.ss_S_ = s_ss

    def clear_k(self):
        self.k = []

    def clear_k_S(self):
        self.k_S = []


class Sergio:

    def __init__(self,
                 rng,
                 number_genes,
                 number_bins,
                 number_sc,
                 noise_params,
                 noise_type,
                 decays,
                 kout=None,
                 kdown=None,
                 dynamics = False,
                 sampling_state = 10,
                 tol = 1e-3,
                 window_length = 100,
                 dt = 0.01,
                 optimize_sampling = False,
                 safety_steps=0,
                 bifurcation_matrix = None,
                 noise_params_splice = None,
                 noise_type_splice = None,
                 splice_ratio = 4, dt_splice = 0.01,
                 migration_rate = None):
        """
        Noise is a gaussian white noise process with zero mean and finite variance.
        noise_params: The amplitude of noise in CLE. This can be a scalar to use
        for all genes or an array with the same size as number_genes.
        Tol: p-Value threshold above which convergence is reached
        window_length: length of non-overlapping window (# time-steps) that is used to realize convergence
        dt: time step used in  CLE
        noise_params and decays: Could be an array of length number_genes, or single value to use the same value for all genes
        number_sc: number of single cells for which expression is simulated
        sampling_state (>=1): single cells are sampled from sampling_state * number_sc steady-state steps
        optimize_sampling: useful for very large graphs. If set True, may help finding a more optimal sampling_state and so may ignore the input sampling_state
        noise_type: We consider three types of noise, 'sp': a single intrinsic noise is associated to production process, 'spd': a single intrinsic noise is associated to both
        production and decay processes, 'dpd': two independent intrinsic noises are associated to production and decay processes
        dynamics: whether simulate splicing or not
        bifurcation_matrix: is a numpy array (nBins_ * nBins) of <1 values; bifurcation_matrix[i,j] indicates whether cell type i differentiates to type j or not. Its value indicates the rate of transition. If dynamics == True, this matrix should be specified
        noise_params_splice: Same as "noise_params" but for splicing. if not specified, the same noise params as pre-mRNA is used
        noise_type_splice: Same as "noise_type" but for splicing. if not specified, the same noise type as pre-mRNA is used
        splice_ratio: it shows the relative amount of spliced mRNA to pre-mRNA (at steady-state) and therefore tunes the decay rate of spliced mRNA as a function of unspliced mRNA. Could be an array of length number_genes, or single value to use the same value for all genes
        dt_splice = time step for integrating splice SDE


        Note1: It's assumed that no two or more bins differentiate into the same new bin i.e. every bin has either 0 or 1 parent bin
        Note2: differentitation rates (e.g. type1 -> type2) specified in bifurcation_matrix specifies the percentage of cells of type2 that are at the vicinity of type1
        """

        self.rng = rng

        self.nGenes_ = number_genes
        self.nBins_ = number_bins
        self.nSC_ = number_sc
        self.sampling_state_ = sampling_state
        self.tol_ = tol
        self.winLen_ = window_length
        self.dt_ = dt
        self.optimize_sampling_ = optimize_sampling
        self.safety_steps = safety_steps
        self.level2verts_ = {}
        self.gID_to_level_and_idx = {} # This dictionary gives the level and idx in self.level2verts_ of a given gene ID
        self.binDict = {} # This maps bin ID to list of gene objects in that bin; only used for dynamics simulations
        self.maxLevels_ = 0
        self.init_concs_ = np.zeros((number_genes, number_bins))
        self.meanExpression = -1 * np.ones((number_genes, number_bins))
        self.noiseType_ = noise_type
        self.dyn_ = dynamics
        self.nConvSteps = np.zeros(number_bins) # This holds the number of simulated steps till convergence

        if kout is None:
            kout = np.zeros(number_genes).astype(bool)
        self.kout = kout

        if kdown is None:
            kdown = np.zeros(number_genes).astype(bool)
        self.kdown = kdown

        ############
        # This graph stores for each vertex: parameters(interaction
        # parameters for non-master regulators and production rates for master
        # regulators), tragets, regulators and level
        ############
        self.graph_ = {}

        if np.isscalar(noise_params):
            self.noiseParamsVector_ = np.repeat(noise_params, number_genes)
        elif np.shape(noise_params)[0] == number_genes:
            self.noiseParamsVector_ = noise_params
        else:
            print ("Error: expect one noise parameter per gene")


        if np.isscalar(decays) == 1:
            self.decayVector_ = np.repeat(decays, number_genes)
        elif np.shape(decays)[0] == number_genes:
            self.decayVector_ = decays
        else:
            print ("Error: expect one decay parameter per gene")
            sys.exit()


    def custom_graph(self, *, g, k, b, hill):
        """
        Prepare custom graph model and coefficients
        Args:
            g: [nGenes_, nGenes_] graph
            k: [nGenes_, nGenes_] interaction coeffs
            b: [nGenes_, nBins_] basal reproduction rate of source nodes (master regulators)
            hill: [nGenes_, nGenes_] hill coefficients of nonlinear interactions
        """

        # check inputs
        assert g.shape == k.shape
        assert g.shape == hill.shape
        assert g.shape[0] == self.nGenes_
        assert g.shape[1] == self.nGenes_
        assert b.shape[0] == self.nGenes_
        assert b.shape[1] == self.nBins_
        assert np.allclose(g[np.diag_indices(g.shape[0])], 0.0), f"No self loops allowed"

        # following steps of original function
        for i in range(self.nGenes_):
            self.graph_[i] = {}
            self.graph_[i]['targets'] = []


        self.master_regulators_idx_ = set()

        for j in range(self.nGenes_):

            is_parent = g[:, j]

            # master regulator (no parents)
            if is_parent.sum() == 0:

                self.master_regulators_idx_.add(j)
                self.graph_[j]['rates'] = b[j]
                self.graph_[j]['regs'] = []
                self.graph_[j]['level'] = -1

            # regular gene (target)
            else:

                currInteraction = []
                currParents = []
                for u in np.where(is_parent == 1)[0]:
                    currInteraction.append((u, k[u, j], hill[u, j], 0))  # last zero shows half-response, it is modified in another method
                    currParents.append(u)
                    self.graph_[u]['targets'].append(j)

                self.graph_[j]['params'] = currInteraction
                self.graph_[j]['regs'] = currParents
                self.graph_[j]['level'] = -1  # will be modified later


        self.find_levels_(self.graph_)

    def find_levels_ (self, graph):
        """
        # This is a helper function that takes a graph and assigns layer to all
        # verticies. It uses longest path layering algorithm from
        # Hierarchical Graph Drawing by Healy and Nikolovself. A bottom-up
        # approach is implemented to optimize simulator run-time. Layer zero is
        # the last layer for which expression are simulated
        # U: verticies with an assigned layer
        # Z: vertizies assigned to a layer below the current layer
        # V: set of all verticies (genes)

        This also sets a dictionary that maps a level to a matrix (in form of python list)
        of all genes in that level versus all bins

        Note to self:
        This is like DFS topsort, but compressing the length of levels as much as possible
        Essentially, root nodes have the highest level (to be simulated first) and sink nodes have level 0,
        and any node upstream of a node has higher level
        Sets:
            level2verts_
                {l: [M, bins] where M is number of genes in on level l}

            gID_to_level_and_idx
                {v: (level, j) where j in index of vertex v in `level2verts[level]` for vertex v in graph
        """

        U = set()
        Z = set()
        V = set(graph.keys())

        currLayer = 0
        self.level2verts_[currLayer] = []
        idx = 0

        while U != V:
            currVerts = set(filter(lambda v: set(graph[v]['targets']).issubset(Z), V-U))

            for v in currVerts:
                graph[v]['level'] = currLayer
                U.add(v)
                if {v}.issubset(self.master_regulators_idx_):
                    allBinList = [Gene(v,'MR', i) for i in range(self.nBins_)]
                    self.level2verts_[currLayer].append(allBinList)
                    self.gID_to_level_and_idx[v] = (currLayer, idx)
                    idx += 1
                else:
                    allBinList = [Gene(v,'T', i) for i in range(self.nBins_)]
                    self.level2verts_[currLayer].append(allBinList)
                    self.gID_to_level_and_idx[v] = (currLayer, idx)
                    idx += 1

            currLayer += 1
            Z = Z.union(U)
            self.level2verts_[currLayer] = []
            idx = 0

        self.level2verts_.pop(currLayer)
        self.maxLevels_ = currLayer - 1

        if not self.dyn_:
            self.set_scIndices_()

    def set_scIndices_ (self):
        """
        # First updates sampling_state_ if optimize_sampling_ is set True: to optimize run time,
        run for less than 30,000 steps in first level
        # Set the single cell indices that are sampled from steady-state steps
        # Note that sampling should be performed from the end of Concentration list
        # Note that this method should run after building graph(and layering) and should
        be run just once!
        """

        if self.optimize_sampling_:
            state = np.true_divide(30000 - self.safety_steps * self.maxLevels_, self.nSC_)
            if state < self.sampling_state_:
                self.sampling_state_ = state

        # time indices when to collect "single-cell" expression snapshots
        self.scIndices_ = self.rng.integers(low = - self.sampling_state_ * self.nSC_, high = 0, size = self.nSC_)

    def calculate_required_steps_(self, level):
        """
        # Calculates the number of required simulation steps after convergence at each level.
        # safety_steps: estimated number of steps required to reach convergence (same), although it is not neede!
        """
        #TODO: remove this safety step

        # Note to self: as safety measure leaving this safety step to double check
        # that knockouts/knockdowns have reached steady-state
        # however, we do initialize the concentrations correctly so should be fine without it

        return self.sampling_state_ * self.nSC_ + level * self.safety_steps

    def calculate_half_response_(self, level):
        """
        Calculates the half response for all interactions between previous layer
        and current layer
        """

        currGenes = self.level2verts_[level]

        for g in currGenes: # g is list of all bins for a single gene
            c = 0
            if g[0].Type == 'T':
                for interTuple in self.graph_[g[0].ID]['params']:
                    regIdx = interTuple[0]
                    meanArr = self.meanExpression[regIdx]

                    if set(meanArr) == set([-1]):
                        print ("Error: Something's wrong in either layering or simulation. Expression of one or more genes in previous layer was not modeled.")
                        sys.exit()

                    self.graph_[g[0].ID]['params'][c] = (self.graph_[g[0].ID]['params'][c][0], self.graph_[g[0].ID]['params'][c][1], self.graph_[g[0].ID]['params'][c][2], np.mean(meanArr))
                    c += 1
            #Else: g is a master regulator and does not need half response

    def hill_(self, reg_conc, half_response, coop_state, repressive = False):
        """
        So far, hill function was used in the code to model 1 interaction at a time.
        So the inputs are single values instead of list or array. Also, it models repression based on this assumption.
        if decided to make it work on multiple interaction, repression should be taken care as well.
        """
        if reg_conc == 0:
            if repressive:
                return 1
            else:
                return 0
        else:
            if repressive:
                return 1 - np.true_divide(np.power(reg_conc, coop_state), (np.power(half_response, coop_state) + np.power(reg_conc, coop_state)) )
            else:
                return np.true_divide(np.power(reg_conc, coop_state), (np.power(half_response, coop_state) + np.power(reg_conc, coop_state)) )

    def init_gene_bin_conc_ (self, level):
        """
        Initilizes the concentration of all genes in the input level

        Note: calculate_half_response_ should be run before this method
        """

        currGenes = self.level2verts_[level]
        for g in currGenes:
            gID = g[0].ID

            # adjust initial concentratino for possible knockout/knockdown experiment
            if self.kout[gID]:
                assert not self.kdown[gID], "Cannot knockout and knockdown the same gene"
                interv_factor = 0.0
            elif self.kdown[gID]:
                interv_factor = 0.5
            else:
                interv_factor = 1.0

            # initialize at expected steady-state
            if g[0].Type == 'MR':
                allBinRates = self.graph_[gID]['rates']

                for bIdx, rate in enumerate(allBinRates):
                    g[bIdx].append_Conc(np.true_divide(interv_factor * rate, self.decayVector_[g[0].ID]))

            else:
                params = self.graph_[g[0].ID]['params']

                for bIdx in range(self.nBins_):
                    rate = 0
                    for interTuple in params:
                        meanExp = self.meanExpression[interTuple[0], bIdx]
                        rate += np.abs(interTuple[1]) * self.hill_(meanExp, interTuple[3], interTuple[2], interTuple[1] < 0)

                    g[bIdx].append_Conc(np.true_divide(interv_factor * rate, self.decayVector_[g[0].ID]))

    def calculate_prod_rate_(self, bin_list, level):
        """
        calculates production rates for the input list of gene objects in different bins but all associated to a single gene ID
        """
        type = bin_list[0].Type

        if (type == 'MR'):
            rates = self.graph_[bin_list[0].ID]['rates']
            return np.array([rates[gb.binID] for gb in bin_list])

        else:
            params = self.graph_[bin_list[0].ID]['params']
            Ks = [np.abs(t[1]) for t in params]
            regIndices = [t[0] for t in params]
            binIndices = [gb.binID for gb in bin_list]
            currStep = bin_list[0].simulatedSteps_
            lastLayerGenes = np.copy(self.level2verts_[level + 1])
            hillMatrix = np.zeros((len(regIndices), len(binIndices)))

            for tupleIdx, rIdx in enumerate(regIndices):
                regGeneLevel = self.gID_to_level_and_idx[rIdx][0]
                regGeneIdx = self.gID_to_level_and_idx[rIdx][1]
                regGene_allBins = self.level2verts_[regGeneLevel][regGeneIdx]
                for colIdx, bIdx in enumerate(binIndices):
                    hillMatrix[tupleIdx, colIdx] = self.hill_(regGene_allBins[bIdx].Conc[currStep], params[tupleIdx][3], params[tupleIdx][2], params[tupleIdx][1] < 0)

            return np.matmul(Ks, hillMatrix)


    def CLE_simulator_(self, level):

        self.calculate_half_response_(level)
        self.init_gene_bin_conc_(level)
        nReqSteps = self.calculate_required_steps_(level)

        # list of lists of genes at the current level
        # each inner list contains `nBins` elements, representing expressions of genes of a given cell type (bin) at time t
        # level2verts_: {level: [[...], ...., [...]]}
        sim_set = np.copy(self.level2verts_[level]).tolist()

        # simulate expression of each gene at this level (vectorized for all bins)
        while sim_set != []:

            delIndicesGenes = []

            # g: list of gene objects of length 'nBins'
            for gi, bin_list in enumerate(sim_set):

                # gene id (row/col in adjacency matrix)
                gID = bin_list[0].ID

                # level in graph, index in list of expressions per bin (same as gi)
                gLevel, gIDX = self.gID_to_level_and_idx[gID]
                assert level == gLevel, "Levels should match"
                assert gi == gIDX, "index in gene-bin matrix should match"

                # [nBins,] current expressions
                currExp = np.array([gb.Conc[-1] for gb in bin_list])

                # [nBins,] production rate of gene given parents
                # if knocked out, set production rate to 0
                # if knocked down, multiply production rate by 0.5
                if self.kout[gID]:
                    prod_rate = np.zeros(len(currExp))
                elif self.kdown[gID]:
                    prod_rate = 0.5 * self.calculate_prod_rate_(bin_list, level)
                else:
                    prod_rate = self.calculate_prod_rate_(bin_list, level)

                # [nBins,] decay rate
                decay = np.multiply(self.decayVector_[gID], currExp)

                # sample noise
                if self.noiseType_ == 'sp':
                    # This notation is inconsistent with our formulation, dw should
                    # include dt^0.5 as well, but here we multipy dt^0.5 later
                    # [nBins, ]
                    dw = self.rng.normal(size = len(currExp))
                    amplitude = np.multiply (self.noiseParamsVector_[gID] , np.power(prod_rate, 0.5))
                    noise = np.multiply(amplitude, dw)

                elif self.noiseType_ == "spd":
                    # [nBins, ]
                    dw = self.rng.normal(size = len(currExp))
                    amplitude = np.multiply (self.noiseParamsVector_[gID] , np.power(prod_rate, 0.5) + np.power(decay, 0.5))
                    noise = np.multiply(amplitude, dw)


                elif self.noiseType_ == "dpd":
                    # [nBins, ]
                    dw_p = self.rng.normal(size = len(currExp))
                    dw_d = self.rng.normal(size = len(currExp))

                    amplitude_p = np.multiply (self.noiseParamsVector_[gID] , np.power(prod_rate, 0.5))
                    amplitude_d = np.multiply (self.noiseParamsVector_[gID] , np.power(decay, 0.5))

                    noise = np.multiply(amplitude_p, dw_p) + np.multiply(amplitude_d, dw_d)

                else:
                    raise KeyError(f"Unknown noise type {self.noiseType_}")

                # [nBins,] change in expression per bin
                dxdt = self.dt_ * (prod_rate - decay) + np.power(self.dt_, 0.5) * noise

                # update expression for each bin
                delIndices = []
                for bIDX, gObj in enumerate(bin_list):

                    # append new concentration level to list of expressions in bin
                    binID = gObj.binID
                    gObj.append_Conc(gObj.Conc[-1] + dxdt[bIDX])
                    gObj.incrementStep()

                    # check whether we collected enough samples
                    if len(gObj.Conc) == nReqSteps:
                        # if so, extract and save expressions at preset time snapshots
                        gObj.set_scExpression(self.scIndices_)
                        self.meanExpression [gID, binID] = np.mean(gObj.scExpression)
                        self.level2verts_[level][gIDX][binID] = gObj
                        delIndices.append(bIDX)

                # remove bins to be simulated when they are done
                sim_set[gi] = [i for j, i in enumerate(bin_list) if j not in delIndices]

                if sim_set[gi] == []:
                    delIndicesGenes.append(gi)

            # remove genes to be simulated if done
            sim_set = [i for j, i in enumerate(sim_set) if j not in delIndicesGenes]


    def simulate(self):
        for level in range(self.maxLevels_, -1, -1):
            self.CLE_simulator_(level)


    def getExpressions(self):
        ret = np.zeros((self.nBins_, self.nGenes_, self.nSC_))
        for l in range(self.maxLevels_ + 1):
            currGeneBins = self.level2verts_[l]
            for g in currGeneBins:
                gIdx = g[0].ID

                for gb in g:
                    ret[gb.binID, gIdx, :] = gb.scExpression

        return ret

    """""""""""""""""""""""""""""""""""""""
    "" This part is to add technical noise
    """""""""""""""""""""""""""""""""""""""
    def outlier_effect(self, scData, outlier_prob, mean, scale):
        """
        Args:
            scData: shape [#cell_types, #genes, #cells_per_type].
        """
        out_indicator = self.rng.binomial(n = 1, p = outlier_prob, size = self.nGenes_)
        outlierGenesIndx = np.where(out_indicator == 1)[0]
        numOutliers = len(outlierGenesIndx)

        #### generate outlier factors ####
        outFactors = self.rng.lognormal(mean = mean, sigma = scale, size = numOutliers)
        ##################################

        scData = np.concatenate(scData, axis = 1)
        for i, gIndx in enumerate(outlierGenesIndx):
            scData[gIndx,:] = scData[gIndx,:] * outFactors[i]

        # return np.split(scData, self.nBins_, axis = 1) # BUG in original code; should return same shape as input
        return np.stack(np.split(scData, self.nBins_, axis = 1))


    def lib_size_effect(self, scData, mean, scale):
        """
        This functions adjusts the mRNA levels in each cell seperately to mimic
        the library size effect. To adjust mRNA levels, cell-specific factors are sampled
        from a log-normal distribution with given mean and scale.

        scData: the simulated data representing mRNA levels (concentrations);
        np.array (#bins * #genes * #cells)

        mean: mean for log-normal distribution

        var: var for log-normal distribution

        returns libFactors ( np.array(nBin, nCell) )
        returns modified single cell data ( np.array(nBin, nGene, nCell) )
        """

        #TODO make sure that having bins does not intefere with this implementation
        ret_data = []

        libFactors = self.rng.lognormal(mean = mean, sigma = scale, size = (self.nBins_, self.nSC_))
        for binExprMatrix, binFactors in zip(scData, libFactors):
            normalizFactors = np.sum(binExprMatrix, axis = 0 )
            binFactors = binFactors / np.where(normalizFactors == 0.0, 1.0, normalizFactors)
            binFactors = binFactors.reshape(1, self.nSC_)
            binFactors = np.repeat(binFactors, self.nGenes_, axis = 0)

            ret_data.append(np.multiply(binExprMatrix, binFactors))


        return libFactors, np.array(ret_data)


    def dropout_indicator(self, scData, shape = 1, percentile = 65):
        """
        This is similar to Splat package

        Input:
        scData can be the output of simulator or any refined version of it
        (e.g. with technical noise)

        shape: the shape of the logistic function

        percentile: the mid-point of logistic functions is set to the given percentile
        of the input scData

        returns: np.array containing binary indactors showing dropouts
        """
        scData = np.array(scData)
        scData_log = np.log(np.add(scData,1))
        log_mid_point = np.percentile(scData_log, percentile)
        prob_ber = np.true_divide (1, 1 + np.exp( -1*shape * (scData_log - log_mid_point) ))

        binary_ind = self.rng.binomial( n = 1, p = prob_ber)

        return binary_ind

    def convert_to_UMIcounts (self, scData):
        """
        Input: scData can be the output of simulator or any refined version of it
        (e.g. with technical noise)
        """
        return self.rng.poisson(scData)




import inspect
import jax
import jax.numpy as jnp
import jax.random as random
from jax.scipy.special import logsumexp

import haiku as hk
import functools

from avici.utils.data_jax import jax_get_train_x, jax_get_x


def layer_norm(*, axis, name=None):
    return hk.LayerNorm(axis=axis, create_scale=True, create_offset=True, name=name)


def set_diagonal(arr, val):
    n_vars = arr.shape[-1]
    return arr.at[..., jnp.arange(n_vars), jnp.arange(n_vars)].set(val)


class BaseModel(hk.Module):

    def __init__(self,
                 layers=8,
                 dim=128,
                 key_size=32,
                 num_heads=8,
                 widening_factor=4,
                 dropout=0.1,
                 out_dim=None,
                 logit_bias_init=-3.0,
                 cosine_temp_init=0.0,
                 ln_axis=-1,
                 name="BaseModel",
                 ):
        super().__init__(name=name)
        self.dim = dim
        self.out_dim = out_dim or dim
        self.layers = 2 * layers
        self.dropout = dropout
        self.ln_axis = ln_axis
        self.widening_factor = widening_factor
        self.num_heads = num_heads
        self.key_size = key_size
        self.logit_bias_init = logit_bias_init
        self.cosine_temp_init = cosine_temp_init
        self.w_init = hk.initializers.VarianceScaling(2.0, "fan_in", "uniform")  # kaiming uniform


    def __call__(self, x, is_training: bool):
        dropout_rate = self.dropout if is_training else 0.0
        z = hk.Linear(self.dim)(x)

        for _ in range(self.layers):
            # mha
            q_in = layer_norm(axis=self.ln_axis)(z)
            k_in = layer_norm(axis=self.ln_axis)(z)
            v_in = layer_norm(axis=self.ln_axis)(z)
            z_attn = hk.MultiHeadAttention(
                num_heads=self.num_heads,
                key_size=self.key_size,
                w_init_scale=2.0,
                model_size=self.dim,
            )(q_in, k_in, v_in)
            z = z + hk.dropout(hk.next_rng_key(), dropout_rate, z_attn)

            # ffn
            z_in = layer_norm(axis=self.ln_axis)(z)
            z_ffn = hk.Sequential([
                hk.Linear(self.widening_factor * self.dim, w_init=self.w_init),
                jax.nn.relu,
                hk.Linear(self.dim, w_init=self.w_init),
            ])(z_in)
            z = z + hk.dropout(hk.next_rng_key(), dropout_rate, z_ffn)

            # flip N and d axes
            z = jnp.swapaxes(z, -3, -2)

        z = layer_norm(axis=self.ln_axis)(z)
        assert z.shape[-2] == x.shape[-2] and z.shape[-3] == x.shape[-3], "Do we have an odd number of layers?"

        # [..., n_vars, dim]
        z = jnp.max(z, axis=-3)

        # u, v dibs embeddings for edge probabilities
        u = hk.Sequential([
            layer_norm(axis=self.ln_axis),
            hk.Linear(self.out_dim, w_init=self.w_init),
        ])(z)
        v = hk.Sequential([
            layer_norm(axis=self.ln_axis),
            hk.Linear(self.out_dim, w_init=self.w_init),
        ])(z)

        # edge logits
        # [..., n_vars, dim], [..., n_vars, dim] -> [..., n_vars, n_vars]
        u /= jnp.linalg.norm(u, axis=-1, ord=2, keepdims=True)
        v /= jnp.linalg.norm(v, axis=-1, ord=2, keepdims=True)
        logit_ij = jnp.einsum("...id,...jd->...ij", u, v)
        temp = hk.get_parameter("learned_temp", (1, 1, 1), logit_ij.dtype,
                                init=hk.initializers.Constant(self.cosine_temp_init)).squeeze()
        logit_ij *= jnp.exp(temp)
        logit_ij_bias = hk.get_parameter("final_matrix_bias", (1, 1, 1), logit_ij.dtype,
                                         init=hk.initializers.Constant(self.logit_bias_init)).squeeze()
        logit_ij += logit_ij_bias

        assert logit_ij.shape[-1] == x.shape[-2] and logit_ij.shape[-2] == x.shape[-2]
        return logit_ij



class InferenceModel:

    def __init__(self, *,
                 model_class,
                 model_kwargs,
                 train_p_obs_only=0.0,
                 acyclicity=None,
                 acyclicity_pow_iters=10,
                 mask_diag=True,
                 ):

        self._train_p_obs_only = jnp.array(train_p_obs_only)
        self._acyclicity_weight = acyclicity
        self._acyclicity_power_iters = acyclicity_pow_iters
        self.mask_diag = mask_diag

        # filter deprecated network kwargs
        sig = inspect.signature(model_class.__init__).parameters
        deprec = list(filter(lambda key: key not in sig, model_kwargs.keys()))
        for k in deprec:
            del model_kwargs[k]
            # print(f"Ignoring deprecated kwarg `{k}` loaded from `model_kwargs` in checkpoint")

        # init forward pass transform
        self.net = hk.transform(lambda *args: model_class(**model_kwargs)(*args))


    @functools.partial(jax.jit, static_argnums=(0, 1))
    def sample_graphs(self, n_samples, params, rng, x):
        """
        Args:
            n_samples: number of samples
            params: hk.Params
            rng
            x: [..., N, d, 2]
            is_count_data [...] bool

        Returns:
            graph samples of shape [..., n_samples, d, d]
        """
        # [..., d, d]
        is_training = False
        logits = self.net.apply(params, rng, x, is_training)
        prob1 = jax.nn.sigmoid(logits)

        # sample graphs
        # [..., n_samples, d, d]
        key, subk = random.split(rng)
        samples = jnp.moveaxis(random.bernoulli(subk, prob1, shape=(n_samples,) + prob1.shape), 0, -3).astype(jnp.int32)
        if self.mask_diag:
            samples = set_diagonal(samples, 0.0)

        return samples


    @functools.partial(jax.jit, static_argnums=(0, 4))
    def infer_edge_logprobs(self, params, rng, x, is_training: bool):
        """
        Args:
            params: hk.Params
            rng
            x: [..., N, d, 2]
            is_training
            is_count_data [...] bool

        Returns:
            logprobs of graph adjacency matrix prediction of shape [..., d, d]
        """
        # [..., d, d]
        logits = self.net.apply(params, rng, x, is_training)
        logp_edges = jax.nn.log_sigmoid(logits)
        if self.mask_diag:
            logp_edges = set_diagonal(logp_edges, -jnp.inf)

        return logp_edges


    def infer_edge_probs(self, params, x):
        """
        For test time inference

        Args:
            params: hk.Params
            x: [..., N, d, 1]
            is_count_data [...] bool

        Returns:
            probabilities of graph adjacency matrix prediction of shape [..., d, d]
        """
        is_training_, dummy_rng_ = False, random.PRNGKey(0) # assume test time
        logp_edges = self.infer_edge_logprobs(params, dummy_rng_, x, is_training_)
        p_edges = jnp.exp(logp_edges)
        return p_edges


    def exp_matmul(self, _logmat, _vec, _axis):
        """
        Matrix-vector multiplication with matrix in log-space
        """
        if _axis == -1:
            _ax_unsqueeze, _ax_sum = -2, -1
        elif _axis == -2:
            _ax_unsqueeze, _ax_sum = -1, -2
        else:
            raise ValueError(f"invalid axis inside exp_matmul")

        _logvec, _logvec_sgn = logsumexp(_logmat, b=jnp.expand_dims(_vec, axis=_ax_unsqueeze),
                                         axis=_ax_sum, return_sign=True)
        return _logvec_sgn * jnp.exp(_logvec)


    def acyclicity_spectral_log(self, logmat, key, power_iterations):
        """
        No Bears acyclicity constraint by
        https://psb.stanford.edu/psb-online/proceedings/psb20/Lee.pdf

        Performed in log-space
        """

        # left/right power iteration
        key, subk = random.split(key)
        u = random.normal(subk, shape=logmat.shape[:-1])
        key, subk = random.split(key)
        v = random.normal(subk, shape=logmat.shape[:-1])

        for t in range(power_iterations):
            # u_new = jnp.einsum('...i,...ij->...j', u, mat)
            u_new = self.exp_matmul(logmat, u, -2) # u @ exp(mat)

            # v_new = jnp.einsum('...ij,...j->...i', mat, v)
            v_new = self.exp_matmul(logmat, v, -1) # exp(mat) @ v

            u = u_new / jnp.linalg.norm(u_new, ord=2, axis=-1, keepdims=True)
            v = v_new / jnp.linalg.norm(v_new, ord=2, axis=-1, keepdims=True)

        u = jax.lax.stop_gradient(u)
        v = jax.lax.stop_gradient(v)

        # largest_eigenvalue = (u @ exp(mat) @ v) / u.dot(v)
        largest_eigenvalue = (
                jnp.einsum('...j,...j->...', u, self.exp_matmul(logmat, v, -1)) /
                jnp.einsum('...j,...j->...', u, v)
        )

        return largest_eigenvalue


    """Training"""
    def loss(self, params, dual, key, data, t, is_training: bool):
        # `data` leaves have leading dimension [1, batch_size_device, ...]

        key, subk = random.split(key)
        if is_training:
            x = jax_get_train_x(subk, data, p_obs_only=self._train_p_obs_only)
        else:
            x = jax_get_x(data)

        n_vars = data["g"].shape[-1]

        ### inference model q(G | D)
        # [..., n_observations, d, 2] --> [..., d, d]
        key, subk = random.split(key)
        logits = self.net.apply(params, subk, x, is_training)

        # get logits [..., d, d]
        logp1 = jax.nn.log_sigmoid(  logits)
        logp0 = jax.nn.log_sigmoid(- logits)

        # labels [..., d, d]
        y_soft = data["g"]

        # mean over edges and skip diagonal (no self-loops)
        # [...]
        loss_eltwise = - (y_soft * logp1 + (1 - y_soft) * logp0)
        if self.mask_diag:
            batch_loss = set_diagonal(loss_eltwise, 0.0).sum((-1, -2)) / (n_vars * (n_vars - 1))
        else:
            batch_loss = loss_eltwise.sum((-1, -2)) / (n_vars * n_vars)

        # [] scalar
        loss_raw = batch_loss.mean() # mean over all available batch dims

        ### acyclicity
        key, subk = random.split(key)
        if self._acyclicity_weight is not None:
            # [..., d, d]
            if self.mask_diag:
                logp_edges = set_diagonal(logp1, -jnp.inf)
            else:
                logp_edges = logp1

            # [...]
            spectral_radii = self.acyclicity_spectral_log(logp_edges, subk, power_iterations=self._acyclicity_power_iters)

            # [] scalars
            ave_acyc_penalty = spectral_radii.mean()
            wgt_acyc_penalty = self._acyclicity_weight(ave_acyc_penalty, t, dual)

        else:
            # [] scalars
            ave_acyc_penalty = jnp.array(0.0)
            wgt_acyc_penalty = jnp.array(0.0)

        # [] scalar
        loss = loss_raw + wgt_acyc_penalty
        aux = {
            "loss_raw": loss_raw,
            "acyc": ave_acyc_penalty,
            "wgt_acyc": wgt_acyc_penalty,
            "mean_z_norm": jnp.abs(logits).mean(),
        }
        return loss, aux
